"""CPU profiling collector."""
from __future__ import absolute_import

import collections
import sys
import threading
import weakref

from ddtrace import compat
from ddtrace.profiling import _attr
from ddtrace.profiling import _periodic
from ddtrace.profiling import _nogevent
from ddtrace.profiling import collector
from ddtrace.profiling import event
from ddtrace.profiling.collector import _threading
from ddtrace.profiling.collector import _traceback
from ddtrace.utils import formats
from ddtrace.vendor import attr
from ddtrace.vendor import six


# NOTE: Do not use LOG here. This code runs under a real OS thread and is unable to acquire any lock of the `logging`
# module without having gevent crashing our dedicated thread.


# Those are special features that might not be available depending on your Python version and platform
FEATURES = {
    "cpu-time": False,
    "stack-exceptions": False,
}

IF UNAME_SYSNAME == "Linux":
    FEATURES['cpu-time'] = True

    from posix.time cimport timespec, clock_gettime
    from posix.types cimport clockid_t
    from cpython.exc cimport PyErr_SetFromErrno

    cdef extern from "<pthread.h>":
        # POSIX says this might be a struct, but CPython relies on it being an unsigned long.
        # We should be defining pthread_t here like this:
        # ctypedef unsigned long pthread_t
        # but e.g. musl libc defines pthread_t as a struct __pthread * which breaks the arithmetic Cython
        # wants to do.
        # We pay this with a warning at compilation time, but it works anyhow.
        int pthread_getcpuclockid(unsigned long thread, clockid_t *clock_id)

    cdef p_pthread_getcpuclockid(tid):
        cdef clockid_t clock_id
        if pthread_getcpuclockid(tid, &clock_id) == 0:
            return clock_id
        PyErr_SetFromErrno(OSError)

    # Python < 3.3 does not have `time.clock_gettime`
    cdef p_clock_gettime_ns(clk_id):
        cdef timespec tp
        if clock_gettime(clk_id, &tp) == 0:
            return int(tp.tv_nsec + tp.tv_sec * 10e8)
        PyErr_SetFromErrno(OSError)

    cdef class _ThreadTime(object):
        cdef dict _last_thread_time

        def __init__(self):
            # This uses a tuple of (pthread_id, thread_native_id) as the key to identify the thread: you'd think using
            # the pthread_t id would be enough, but the glibc reuses the id.
            self._last_thread_time = {}

        # Only used in tests
        def _get_last_thread_time(self):
            return dict(self._last_thread_time)

        def __call__(self, pthread_ids):
            cdef list cpu_times = []
            for pthread_id in pthread_ids:
                # TODO: Use QueryThreadCycleTime on Windows?
                # ⚠ WARNING ⚠
                # `pthread_getcpuclockid` can make Python segfault if the thread is does not exist anymore.
                # In order avoid this, this function must be called with the GIL being held the entire time.
                # This is why this whole file is compiled down to C: we make sure we never release the GIL between
                # calling sys._current_frames() and pthread_getcpuclockid, making sure no thread disappeared.
                try:
                    cpu_time = p_clock_gettime_ns(p_pthread_getcpuclockid(pthread_id))
                except OSError:
                    # Just in case it fails, set it to 0
                    # (Note that glibc never fails, it segfaults instead)
                    cpu_time = 0
                cpu_times.append(cpu_time)

            cdef dict pthread_cpu_time = {}

            # We should now be safe doing more Pythonic stuff and maybe releasing the GIL
            for pthread_id, cpu_time in zip(pthread_ids, cpu_times):
                thread_native_id = _threading.get_thread_native_id(pthread_id)
                key = pthread_id, thread_native_id
                # Do a max(0, …) here just in case the result is < 0:
                # This should never happen, but it can happen if the one chance in a billion happens:
                # - A new thread has been created and has the same native id and the same pthread_id.
                # - We got an OSError with clock_gettime_ns
                pthread_cpu_time[key] = max(0, cpu_time - self._last_thread_time.get(key, cpu_time))
                self._last_thread_time[key] = cpu_time

            # Clear cache
            keys = list(pthread_cpu_time.keys())
            for key in list(self._last_thread_time.keys()):
                if key not in keys:
                    del self._last_thread_time[key]

            return pthread_cpu_time
ELSE:
    cdef class _ThreadTime(object):
        cdef long _last_process_time

        def __init__(self):
            self._last_process_time = compat.process_time_ns()

        def __call__(self, pthread_ids):
            current_process_time = compat.process_time_ns()
            cpu_time = current_process_time - self._last_process_time
            self._last_process_time = current_process_time
            # Spread the consumed CPU time on all threads.
            # It's not fair, but we have no clue which CPU used more unless we can use `pthread_getcpuclockid`
            # Check that we don't have zero thread — _might_ very rarely happen at shutdown
            nb_threads = len(pthread_ids)
            if nb_threads == 0:
                cpu_time = 0
            else:
                cpu_time //= nb_threads
            return {
                (pthread_id, _threading.get_thread_native_id(pthread_id)): cpu_time
                for pthread_id in pthread_ids
            }


@event.event_class
class StackSampleEvent(event.StackBasedEvent):
    """A sample storing executions frames for a thread."""

    # Wall clock
    wall_time_ns = attr.ib(default=0)
    # CPU time in nanoseconds
    cpu_time_ns = attr.ib(default=0)


@event.event_class
class StackExceptionSampleEvent(event.StackBasedEvent):
    """A a sample storing raised exceptions and their stack frames."""

    exc_type = attr.ib(default=None)

from cpython.object cimport PyObject

# The head lock (the interpreter mutex) is only exposed in a data structure in Python ≥ 3.7
IF UNAME_SYSNAME != "Windows" and PY_MAJOR_VERSION >= 3 and PY_MINOR_VERSION >= 7:
    FEATURES['stack-exceptions'] = True

    from cpython cimport PyInterpreterState
    from cpython cimport PyInterpreterState_Head, PyInterpreterState_Next
    from cpython cimport PyInterpreterState_ThreadHead, PyThreadState_Next

    from cpython.pythread cimport (
        PyThread_acquire_lock, PyThread_release_lock,
        WAIT_LOCK,
        PyThread_type_lock,
        PY_LOCK_ACQUIRED
    )

    cdef extern from "<Python.h>":
        # This one is provided as an opaque struct from Cython's cpython/pystate.pxd,
        # but we need to access some of its fields so we redefine it here.
        ctypedef struct PyThreadState:
            unsigned long thread_id
            PyObject* frame

        _PyErr_StackItem * _PyErr_GetTopmostException(PyThreadState *tstate)

        ctypedef struct _PyErr_StackItem:
            PyObject* exc_type
            PyObject* exc_value
            PyObject* exc_traceback

    IF PY_MINOR_VERSION == 7:
        # Python 3.7
        cdef extern from "<internal/pystate.h>":

            cdef struct pyinterpreters:
                PyThread_type_lock mutex

            ctypedef struct _PyRuntimeState:
                pyinterpreters interpreters

            cdef extern _PyRuntimeState _PyRuntime

    ELIF PY_MINOR_VERSION >= 8:
        # Python 3.8
        cdef extern from "<internal/pycore_pystate.h>":

            cdef struct pyinterpreters:
                PyThread_type_lock mutex

            ctypedef struct _PyRuntimeState:
                pyinterpreters interpreters

            cdef extern _PyRuntimeState _PyRuntime

        IF PY_MINOR_VERSION >= 9:
            # Needed for accessing _PyGC_FINALIZED when we build with -DPy_BUILD_CORE
            cdef extern from "<internal/pycore_gc.h>":
                pass
ELSE:
    from cpython.ref cimport Py_DECREF

    cdef extern from "<pystate.h>":
        PyObject* _PyThread_CurrentFrames()


cdef collect_threads(ignore_profiler, thread_time, thread_span_links) with gil:
    cdef dict current_exceptions = {}

    IF UNAME_SYSNAME != "Windows" and PY_MAJOR_VERSION >= 3 and PY_MINOR_VERSION >= 7:
        cdef PyInterpreterState* interp
        cdef PyThreadState* tstate
        cdef _PyErr_StackItem* exc_info
        cdef PyThread_type_lock lmutex = _PyRuntime.interpreters.mutex

        cdef dict running_threads = {}

        # This is an internal lock but we do need it.
        # See https://bugs.python.org/issue1021318
        if PyThread_acquire_lock(lmutex, WAIT_LOCK) == PY_LOCK_ACQUIRED:
            # Do not try to do anything fancy here:
            # Even calling print() will deadlock the program has it will try
            # to lock the GIL and somehow touching this mutex.
            try:
                interp = PyInterpreterState_Head()

                while interp:
                    tstate = PyInterpreterState_ThreadHead(interp)
                    while tstate:
                        # The frame can be NULL
                        if tstate.frame:
                            running_threads[tstate.thread_id] = <object>tstate.frame

                        exc_info = _PyErr_GetTopmostException(tstate)
                        if exc_info and exc_info.exc_type and exc_info.exc_traceback:
                            current_exceptions[tstate.thread_id] = (<object>exc_info.exc_type, <object>exc_info.exc_traceback)

                        tstate = PyThreadState_Next(tstate)

                    interp = PyInterpreterState_Next(interp)
            finally:
                PyThread_release_lock(lmutex)
    ELSE:
        cdef dict running_threads = <dict>_PyThread_CurrentFrames()

        # Now that we own the ref via <dict> casting, we can safely decrease the default refcount
        # so we don't leak the object
        Py_DECREF(running_threads)

    cdef dict cpu_times = thread_time(running_threads.keys())

    return tuple(
        (
            pthread_id,
            native_thread_id,
            _threading.get_thread_name(pthread_id),
            running_threads[pthread_id],
            current_exceptions.get(pthread_id),
            thread_span_links.get_active_leaf_spans_from_thread_id(pthread_id) if thread_span_links else set(),
            cpu_time,
        )
        for (pthread_id, native_thread_id), cpu_time in cpu_times.items()
        if not ignore_profiler or pthread_id not in _periodic.PERIODIC_THREADS
    )



cdef stack_collect(ignore_profiler, thread_time, max_nframes, interval, wall_time, thread_span_links):

    running_threads = collect_threads(ignore_profiler, thread_time, thread_span_links)

    if thread_span_links:
        # FIXME also use native thread id
        thread_span_links.clear_threads(tuple(thread[0] for thread in running_threads))

    stack_events = []
    exc_events = []

    for thread_id, thread_native_id, thread_name, frame, exception, spans, cpu_time in running_threads:
        frames, nframes = _traceback.pyframe_to_frames(frame, max_nframes)
        stack_events.append(
            StackSampleEvent(
                thread_id=thread_id,
                thread_native_id=thread_native_id,
                thread_name=thread_name,
                trace_ids=set(span.trace_id for span in spans),
                nframes=nframes, frames=frames,
                wall_time_ns=wall_time,
                cpu_time_ns=cpu_time,
                sampling_period=int(interval * 1e9),
            ),
        )

        if exception is not None:
            exc_type, exc_traceback = exception
            frames, nframes = _traceback.traceback_to_frames(exc_traceback, max_nframes)
            exc_events.append(
                StackExceptionSampleEvent(
                    thread_id=thread_id,
                    thread_name=thread_name,
                    thread_native_id=thread_native_id,
                    nframes=nframes,
                    frames=frames,
                    sampling_period=int(interval * 1e9),
                    exc_type=exc_type,
                ),
            )

    return stack_events, exc_events


@attr.s(slots=True, eq=False)
class _ThreadSpanLinks(object):

    # Keys is a thread_id
    # Value is a set of weakrefs to spans
    _thread_id_to_spans = attr.ib(factory=lambda: collections.defaultdict(set), repr=False, init=False)
    _lock = attr.ib(factory=_nogevent.Lock, repr=False, init=False)

    def link_span(self, span):
        """Link a span to its running environment.

        Track threads, tasks, etc.
        """
        # Since we're going to iterate over the set, make sure it's locked
        with self._lock:
            self._thread_id_to_spans[_nogevent.thread_get_ident()].add(weakref.ref(span))

    def clear_threads(self, existing_thread_ids):
        """Clear the stored list of threads based on the list of existing thread ids.

        If any thread that is part of this list was stored, its data will be deleted.

        :param existing_thread_ids: A set of thread ids to keep.
        """
        with self._lock:
            # Iterate over a copy of the list of keys since it's mutated during our iteration.
            for thread_id in list(self._thread_id_to_spans.keys()):
                if thread_id not in existing_thread_ids:
                    del self._thread_id_to_spans[thread_id]

    def get_active_leaf_spans_from_thread_id(self, thread_id):
        """Return the latest active spans for a thread.

        In theory this should return a single span, though if multiple children span are active without being finished,
        there can be several spans returned.

        :param thread_id: The thread id.
        :return: A set with the active spans.
        """
        alive_spans = set()

        with self._lock:
            span_list = self._thread_id_to_spans.get(thread_id, ())
            dead_spans = set()
            for span_ref in span_list:
                span = span_ref()
                if span is None:
                    dead_spans.add(span_ref)
                else:
                    alive_spans.add(span)

            # Clean the set from the dead spans
            for dead_span in dead_spans:
                span_list.remove(dead_span)

        # Iterate over a copy so we can modify the original
        for span in alive_spans.copy():
            if not span.finished:
                try:
                    alive_spans.remove(span._parent)
                except KeyError:
                    pass

        return {span for span in alive_spans if not span.finished}


def _default_min_interval_time():
    if six.PY2:
        return 0.01
    return sys.getswitchinterval() * 2


@attr.s(slots=True)
class StackCollector(collector.PeriodicCollector):
    """Execution stacks collector."""
    # This need to be a real OS thread in order to catch
    _real_thread = True
    _interval = attr.ib(factory=_default_min_interval_time, init=False, repr=False)
    # This is the minimum amount of time the thread will sleep between polling interval,
    # no matter how fast the computer is.
    min_interval_time = attr.ib(factory=_default_min_interval_time, init=False)

    max_time_usage_pct = attr.ib(factory=_attr.from_env("DD_PROFILING_MAX_TIME_USAGE_PCT", 2, float))
    nframes = attr.ib(factory=_attr.from_env("DD_PROFILING_MAX_FRAMES", 64, int))
    ignore_profiler = attr.ib(factory=_attr.from_env("DD_PROFILING_IGNORE_PROFILER", True, formats.asbool))
    tracer = attr.ib(default=None)
    _thread_time = attr.ib(init=False, repr=False)
    _last_wall_time = attr.ib(init=False, repr=False)
    _thread_span_links = attr.ib(default=None, init=False, repr=False)

    @max_time_usage_pct.validator
    def _check_max_time_usage(self, attribute, value):
        if value <= 0 or value > 100:
            raise ValueError("Max time usage percent must be greater than 0 and smaller or equal to 100")

    def _init(self):
        self._thread_time = _ThreadTime()
        self._last_wall_time = compat.monotonic_ns()
        if self.tracer is not None:
            self._thread_span_links = _ThreadSpanLinks()
            self.tracer.on_start_span(self._thread_span_links.link_span)

    def start(self):
        # This is split in its own function to ease testing
        self._init()
        super(StackCollector, self).start()

    def stop(self):
        super(StackCollector, self).stop()
        if self.tracer is not None:
            self.tracer.deregister_on_start_span(self._thread_span_links.link_span)

    def _compute_new_interval(self, used_wall_time_ns):
        interval = (used_wall_time_ns / (self.max_time_usage_pct / 100.0)) - used_wall_time_ns
        return max(interval / 1e9, self.min_interval_time)

    def collect(self):
        # Compute wall time
        now = compat.monotonic_ns()
        wall_time = now - self._last_wall_time
        self._last_wall_time = now

        all_events = stack_collect(
            self.ignore_profiler, self._thread_time, self.nframes, self.interval, wall_time, self._thread_span_links,
        )

        used_wall_time_ns = compat.monotonic_ns() - now
        self.interval = self._compute_new_interval(used_wall_time_ns)

        return all_events



import sys
import time

import mysql.connector
from psycopg2 import connect, OperationalError
from cassandra.cluster import Cluster, NoHostAvailable
import rediscluster
import vertica_python
import kombu

from contrib.config import (
    POSTGRES_CONFIG,
    CASSANDRA_CONFIG,
    MYSQL_CONFIG,
    REDISCLUSTER_CONFIG,
    VERTICA_CONFIG,
    RABBITMQ_CONFIG
)


def try_until_timeout(exception):
    """Utility decorator that tries to call a check until there is a
    timeout.  The default timeout is about 20 seconds.

    """
    def wrap(fn):
        def wrapper(*args, **kwargs):
            err = None

            for i in range(100):
                try:
                    fn()
                except exception as e:
                    err = e
                    time.sleep(0.2)
                else:
                    break
            else:
                if err:
                    raise err
        return wrapper
    return wrap


@try_until_timeout(OperationalError)
def check_postgres():
    conn = connect(**POSTGRES_CONFIG)
    try:
        conn.cursor().execute('SELECT 1;')
    finally:
        conn.close()


@try_until_timeout(NoHostAvailable)
def check_cassandra():
    with Cluster(**CASSANDRA_CONFIG).connect() as conn:
        conn.execute('SELECT now() FROM system.local')


@try_until_timeout(Exception)
def check_mysql():
    conn = mysql.connector.connect(**MYSQL_CONFIG)
    try:
        conn.cursor().execute('SELECT 1;')
    finally:
        conn.close()


@try_until_timeout(Exception)
def check_rediscluster():
    test_host = REDISCLUSTER_CONFIG['host']
    test_ports = REDISCLUSTER_CONFIG['ports']
    startup_nodes = [
        {'host': test_host, 'port': int(port)}
        for port in test_ports.split(',')
    ]
    r = rediscluster.StrictRedisCluster(startup_nodes=startup_nodes)
    r.flushall()


@try_until_timeout(Exception)
def check_vertica():
    conn = vertica_python.connect(**VERTICA_CONFIG)
    try:
        conn.cursor().execute('SELECT 1;')
    finally:
        conn.close()


@try_until_timeout(Exception)
def check_rabbitmq():
    url = 'amqp://{user}:{password}@{host}:{port}//'.format(**RABBITMQ_CONFIG)
    conn = kombu.Connection(url)
    try:
        conn.connect()
    finally:
        conn.release()


if __name__ == '__main__':
    check_functions = {
        'cassandra': check_cassandra,
        'postgres': check_postgres,
        'mysql': check_mysql,
        'rediscluster': check_rediscluster,
        'vertica': check_vertica,
        'rabbitmq': check_rabbitmq,
    }
    if len(sys.argv) >= 2:
        for service in sys.argv[1:]:
            check_functions[service]()
    else:
        print('usage: python {} SERVICE_NAME'.format(sys.argv[0]))
        sys.exit(1)



import mock
from ddtrace.internal.runtime.collector import ValueCollector

from tests import BaseTestCase


def mocked_collector(mock_collect, **kwargs):
    collector = ValueCollector(**kwargs)
    collector.collect_fn = mock_collect
    return collector


class TestValueCollector(BaseTestCase):

    def test_default_usage(self):
        mock_collect = mock.MagicMock()
        mock_collect.side_effect = lambda k: [
            ('key1', 'value1'),
            ('key2', 'value2'),
        ]

        vc = mocked_collector(mock_collect)

        self.assertEqual(vc.collect(keys=set(['key1'])), [
            ('key1', 'value1'),
        ])
        mock_collect.assert_called_once()
        mock_collect.assert_called_with(set(['key1']))

        self.assertEqual(mock_collect.call_count, 1,
                         'Collector is not periodic by default')

    def test_enabled(self):
        collect = mock.MagicMock()
        vc = mocked_collector(collect, enabled=False)
        collect.assert_not_called()
        vc.collect()
        collect.assert_not_called()

    def test_periodic(self):
        collect = mock.MagicMock()
        vc = mocked_collector(collect, periodic=True)
        vc.collect()
        self.assertEqual(collect.call_count, 1)
        vc.collect()
        self.assertEqual(collect.call_count, 2)

    def test_not_periodic(self):
        collect = mock.MagicMock()
        vc = mocked_collector(collect)
        collect.assert_not_called()
        vc.collect()
        self.assertEqual(collect.call_count, 1)
        vc.collect()
        self.assertEqual(collect.call_count, 1)
        vc.collect()
        self.assertEqual(collect.call_count, 1)

    def test_required_module(self):
        mock_module = mock.MagicMock()
        mock_module.fn.side_effect = lambda: 'test'
        with self.override_sys_modules(dict(A=mock_module)):
            class AVC(ValueCollector):
                required_modules = ['A']

                def collect_fn(self, keys):
                    a = self.modules.get('A')
                    a.fn()

            vc = AVC()
            vc.collect()
            mock_module.fn.assert_called_once()

    def test_required_module_not_installed(self):
        collect = mock.MagicMock()
        with mock.patch('ddtrace.internal.runtime.collector.log') as log_mock:
            # Should log a warning (tested below)
            vc = mocked_collector(collect, required_modules=['moduleshouldnotexist'])

            # Collect should not be called as the collector should be disabled.
            collect.assert_not_called()
            vc.collect()
            collect.assert_not_called()

        calls = [
            mock.call(
                'Could not import module "%s" for %s. Disabling collector.',
                'moduleshouldnotexist', vc,
            )
        ]
        log_mock.warning.assert_has_calls(calls)

    def test_collected_values(self):
        class V(ValueCollector):
            i = 0

            def collect_fn(self, keys):
                self.i += 1
                return [('i', self.i)]

        vc = V()
        self.assertEqual(vc.collect(), [('i', 1)])
        self.assertEqual(vc.collect(), [('i', 1)])
        self.assertEqual(vc.collect(), [('i', 1)])

    def test_collected_values_periodic(self):
        class V(ValueCollector):
            periodic = True
            i = 0

            def collect_fn(self, keys):
                self.i += 1
                return [('i', self.i)]

        vc = V()
        self.assertEqual(vc.collect(), [('i', 1)])
        self.assertEqual(vc.collect(), [('i', 2)])
        self.assertEqual(vc.collect(), [('i', 3)])



from .config import Config
from .exceptions import ConfigException
from .http import HttpConfig
from .._hooks import Hooks
from .integration import IntegrationConfig

# Default global config
config = Config()

__all__ = [
    "config",
    "Config",
    "ConfigException",
    "HttpConfig",
    "Hooks",
    "IntegrationConfig",
]



from ddtrace.constants import SAMPLING_PRIORITY_KEY, ORIGIN_KEY

from .utils import PyramidTestCase, PyramidBase


def includeme(config):
    pass


class TestPyramid(PyramidTestCase):
    instrument = True

    def test_tween_overridden(self):
        # in case our tween is overriden by the user config we should
        # not log rendering
        self.override_settings({'pyramid.tweens': 'pyramid.tweens.excview_tween_factory'})
        self.app.get('/json', status=200)
        spans = self.tracer.writer.pop()
        assert len(spans) == 0


class TestPyramidDistributedTracingDefault(PyramidBase):
    instrument = True

    def get_settings(self):
        return {}

    def test_distributed_tracing(self):
        # ensure the Context is properly created
        # if distributed tracing is enabled
        headers = {
            'x-datadog-trace-id': '100',
            'x-datadog-parent-id': '42',
            'x-datadog-sampling-priority': '2',
            'x-datadog-origin': 'synthetics',
        }
        self.app.get('/', headers=headers, status=200)
        writer = self.tracer.writer
        spans = writer.pop()
        assert len(spans) == 1
        # check the propagated Context
        span = spans[0]
        assert span.trace_id == 100
        assert span.parent_id == 42
        assert span.get_metric(SAMPLING_PRIORITY_KEY) == 2
        assert span.get_tag(ORIGIN_KEY) == 'synthetics'


class TestPyramidDistributedTracingDisabled(PyramidBase):
    instrument = True

    def get_settings(self):
        return {
            'datadog_distributed_tracing': False,
        }

    def test_distributed_tracing_disabled(self):
        # we do not inherit context if distributed tracing is disabled
        headers = {
            'x-datadog-trace-id': '100',
            'x-datadog-parent-id': '42',
            'x-datadog-sampling-priority': '2',
            'x-datadog-origin': 'synthetics',
        }
        self.app.get('/', headers=headers, status=200)
        writer = self.tracer.writer
        spans = writer.pop()
        assert len(spans) == 1
        # check the propagated Context
        span = spans[0]
        assert span.trace_id != 100
        assert span.parent_id != 42
        assert span.get_metric(SAMPLING_PRIORITY_KEY) != 2
        assert span.get_tag(ORIGIN_KEY) != 'synthetics'



import pytest

from ddtrace import Pin, Tracer
from ddtrace.settings import Config
from ddtrace.contrib import trace_utils

from tests import override_global_config


@pytest.fixture
def config():
    c = Config()
    c._add("myint", dict())
    return c


@pytest.mark.parametrize(
    "pin,config_val,default,global_service,expected",
    [
        (Pin(), None, None, None, None),
        (Pin(), None, None, "global-svc", "global-svc"),
        (Pin(), None, "default-svc", None, "default-svc"),
        # Global service should have higher priority than the integration default.
        (Pin(), None, "default-svc", "global-svc", "global-svc"),
        (Pin(), "config-svc", "default-svc", None, "config-svc"),
        (Pin(service="pin-svc"), None, "default-svc", None, "pin-svc"),
        (Pin(service="pin-svc"), "config-svc", "default-svc", None, "pin-svc"),
    ],
)
def test_int_service(config, pin, config_val, default, global_service, expected):
    if config_val:
        config.myint.service = config_val

    if global_service:
        config.service = global_service

    assert trace_utils.int_service(pin, config.myint, default) == expected


def test_int_service_integration(config):
    pin = Pin()
    tracer = Tracer()
    assert trace_utils.int_service(pin, config.myint) is None

    with override_global_config(dict(service="global-svc")):
        assert trace_utils.int_service(pin, config.myint) is None

        with tracer.trace("something", service=trace_utils.int_service(pin, config.myint)) as s:
            assert s.service == "global-svc"


@pytest.mark.parametrize(
    "pin,config_val,default,expected",
    [
        (Pin(), None, "default-svc", "default-svc"),
        (Pin(), "config-svc", "default-svc", "config-svc"),
        (Pin(service="pin-svc"), None, "default-svc", "pin-svc"),
        (Pin(service="pin-svc"), "config-svc", "default-svc", "pin-svc"),
    ],
)
def test_ext_service(config, pin, config_val, default, expected):
    if config_val:
        config.myint.service = config_val

    assert trace_utils.ext_service(pin, config.myint, default) == expected



from pylons.controllers import WSGIController

from ..lib.helpers import ExceptionWithCodeMethod, get_render_fn


class BaseController(WSGIController):

    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        # WSGIController.__call__ dispatches to the Controller method
        # the request is routed to. This routing information is
        # available in environ['pylons.routes_dict']
        return WSGIController.__call__(self, environ, start_response)


class RootController(BaseController):
    """Controller used for most tests"""

    def index(self):
        return 'Hello World'

    def raise_exception(self):
        raise Exception('Ouch!')

    def raise_wrong_code(self):
        e = Exception('Ouch!')
        e.code = 'wrong formatted code'
        raise e

    def raise_code_method(self):
        raise ExceptionWithCodeMethod('Ouch!')

    def raise_custom_code(self):
        e = Exception('Ouch!')
        e.code = '512'
        raise e

    def render(self):
        render = get_render_fn()
        return render('/template.mako')

    def render_exception(self):
        render = get_render_fn()
        return render('/exception.mako')



# flake8: noqa
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import hello_pb2 as hello__pb2


class HelloStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.SayHello = channel.unary_unary(
        '/helloworld.Hello/SayHello',
        request_serializer=hello__pb2.HelloRequest.SerializeToString,
        response_deserializer=hello__pb2.HelloReply.FromString,
        )
    self.SayHelloTwice = channel.unary_stream(
        '/helloworld.Hello/SayHelloTwice',
        request_serializer=hello__pb2.HelloRequest.SerializeToString,
        response_deserializer=hello__pb2.HelloReply.FromString,
        )
    self.SayHelloRepeatedly = channel.stream_stream(
        '/helloworld.Hello/SayHelloRepeatedly',
        request_serializer=hello__pb2.HelloRequest.SerializeToString,
        response_deserializer=hello__pb2.HelloReply.FromString,
        )
    self.SayHelloLast = channel.stream_unary(
        '/helloworld.Hello/SayHelloLast',
        request_serializer=hello__pb2.HelloRequest.SerializeToString,
        response_deserializer=hello__pb2.HelloReply.FromString,
        )
    self.SayHelloUnknown = channel.unary_unary(
        '/helloworld.Hello/SayHelloUnknown',
        request_serializer=hello__pb2.HelloRequest.SerializeToString,
        response_deserializer=hello__pb2.HelloReply.FromString,
        )


class HelloServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def SayHello(self, request, context):
    """Sends a greeting
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SayHelloTwice(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SayHelloRepeatedly(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SayHelloLast(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SayHelloUnknown(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

def add_HelloServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'SayHello': grpc.unary_unary_rpc_method_handler(
          servicer.SayHello,
          request_deserializer=hello__pb2.HelloRequest.FromString,
          response_serializer=hello__pb2.HelloReply.SerializeToString,
      ),
      'SayHelloTwice': grpc.unary_stream_rpc_method_handler(
          servicer.SayHelloTwice,
          request_deserializer=hello__pb2.HelloRequest.FromString,
          response_serializer=hello__pb2.HelloReply.SerializeToString,
      ),
      'SayHelloRepeatedly': grpc.stream_stream_rpc_method_handler(
          servicer.SayHelloRepeatedly,
          request_deserializer=hello__pb2.HelloRequest.FromString,
          response_serializer=hello__pb2.HelloReply.SerializeToString,
      ),
      'SayHelloLast': grpc.stream_unary_rpc_method_handler(
          servicer.SayHelloLast,
          request_deserializer=hello__pb2.HelloRequest.FromString,
          response_serializer=hello__pb2.HelloReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'helloworld.Hello', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



import pymemcache

from ddtrace.ext import memcached as memcachedx
from ddtrace.pin import Pin, _DD_PIN_NAME, _DD_PIN_PROXY_NAME
from .client import WrappedClient

_Client = pymemcache.client.base.Client


def patch():
    if getattr(pymemcache.client, '_datadog_patch', False):
        return

    setattr(pymemcache.client, '_datadog_patch', True)
    setattr(pymemcache.client.base, 'Client', WrappedClient)

    # Create a global pin with default configuration for our pymemcache clients
    Pin(
        app=memcachedx.SERVICE, service=memcachedx.SERVICE
    ).onto(pymemcache)


def unpatch():
    """Remove pymemcache tracing"""
    if not getattr(pymemcache.client, '_datadog_patch', False):
        return
    setattr(pymemcache.client, '_datadog_patch', False)
    setattr(pymemcache.client.base, 'Client', _Client)

    # Remove any pins that may exist on the pymemcache reference
    setattr(pymemcache, _DD_PIN_NAME, None)
    setattr(pymemcache, _DD_PIN_PROXY_NAME, None)



"""
tests for parsing specs.
"""

from bson.son import SON

from ddtrace.contrib.pymongo.parse import parse_spec


def test_empty():
    cmd = parse_spec(SON([]))
    assert cmd is None


def test_create():
    cmd = parse_spec(SON([('create', 'foo')]))
    assert cmd.name == 'create'
    assert cmd.coll == 'foo'
    assert cmd.tags == {}
    assert cmd.metrics == {}


def test_insert():
    spec = SON([
        ('insert', 'bla'),
        ('ordered', True),
        ('documents', ['a', 'b']),
    ])
    cmd = parse_spec(spec)
    assert cmd.name == 'insert'
    assert cmd.coll == 'bla'
    assert cmd.tags == {'mongodb.ordered': True}
    assert cmd.metrics == {'mongodb.documents': 2}


def test_update():
    spec = SON([
        ('update', u'songs'),
        ('ordered', True),
        ('updates', [
            SON([
                ('q', {'artist': 'Neil'}),
                ('u', {'$set': {'artist': 'Shakey'}}),
                ('multi', True),
                ('upsert', False)
            ])
        ])
    ])
    cmd = parse_spec(spec)
    assert cmd.name == 'update'
    assert cmd.coll == 'songs'
    assert cmd.query == {'artist': 'Neil'}



import threading

from ddtrace.context import Context
from ddtrace.internal.context_manager import DefaultContextManager
from ddtrace.span import Span

from tests import BaseTestCase


class TestDefaultContextManager(BaseTestCase):
    """
    Ensures that a ``ContextManager`` makes the Context
    local to each thread or task.
    """
    def test_get_or_create(self):
        # asking the Context multiple times should return
        # always the same instance
        ctxm = DefaultContextManager()
        assert ctxm.get() == ctxm.get()

    def test_set_context(self):
        # the Context can be set in the current Thread
        ctx = Context()
        ctxm = DefaultContextManager()
        assert ctxm.get() is not ctx

        ctxm.set(ctx)
        assert ctxm.get() is ctx

    def test_multiple_threads_multiple_context(self):
        # each thread should have it's own Context
        ctxm = DefaultContextManager()

        def _fill_ctx():
            ctx = ctxm.get()
            span = Span(tracer=None, name='fake_span')
            ctx.add_span(span)
            assert 1 == len(ctx._trace)

        threads = [threading.Thread(target=_fill_ctx) for _ in range(100)]

        for t in threads:
            t.daemon = True
            t.start()

        for t in threads:
            t.join()

        # the main instance should have an empty Context
        # because it has not been used in this thread
        ctx = ctxm.get()
        assert 0 == len(ctx._trace)

    def test_reset_context_manager(self):
        ctxm = DefaultContextManager()
        ctx = ctxm.get()

        # new context manager should not share same context
        ctxm = DefaultContextManager()
        assert ctxm.get() is not ctx



from requests_mock import Adapter

from ddtrace import config

from ... import TracerTestCase
from .test_requests import BaseRequestTestCase


class TestRequestsDistributed(BaseRequestTestCase, TracerTestCase):
    def headers_here(self, tracer, request, root_span):
        # Use an additional matcher to query the request headers.
        # This is because the parent_id can only been known within such a callback,
        # as it's defined on the requests span, which is not available when calling register_uri
        headers = request.headers
        assert 'x-datadog-trace-id' in headers
        assert 'x-datadog-parent-id' in headers
        assert str(root_span.trace_id) == headers['x-datadog-trace-id']
        req_span = root_span.context.get_current_span()
        assert 'requests.request' == req_span.name
        assert str(req_span.span_id) == headers['x-datadog-parent-id']
        return True

    def headers_not_here(self, tracer, request):
        headers = request.headers
        assert 'x-datadog-trace-id' not in headers
        assert 'x-datadog-parent-id' not in headers
        return True

    def test_propagation_default(self):
        # ensure by default, distributed tracing is enabled
        adapter = Adapter()
        self.session.mount('mock', adapter)

        with self.tracer.trace('root') as root:
            def matcher(request):
                return self.headers_here(self.tracer, request, root)
            adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
            resp = self.session.get('mock://datadog/foo')
            assert 200 == resp.status_code
            assert 'bar' == resp.text

    def test_propagation_true_global(self):
        # distributed tracing can be enabled globally
        adapter = Adapter()
        self.session.mount('mock', adapter)

        with self.override_config('requests', dict(distributed_tracing=True)):
            with self.tracer.trace('root') as root:
                def matcher(request):
                    return self.headers_here(self.tracer, request, root)
                adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
                resp = self.session.get('mock://datadog/foo')
                assert 200 == resp.status_code
                assert 'bar' == resp.text

    def test_propagation_false_global(self):
        # distributed tracing can be disabled globally
        adapter = Adapter()
        self.session.mount('mock', adapter)

        with self.override_config('requests', dict(distributed_tracing=False)):
            with self.tracer.trace('root'):
                def matcher(request):
                    return self.headers_not_here(self.tracer, request)
                adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
                resp = self.session.get('mock://datadog/foo')
                assert 200 == resp.status_code
                assert 'bar' == resp.text

    def test_propagation_true(self):
        # ensure distributed tracing can be enabled manually
        cfg = config.get_from(self.session)
        cfg['distributed_tracing'] = True
        adapter = Adapter()
        self.session.mount('mock', adapter)

        with self.tracer.trace('root') as root:
            def matcher(request):
                return self.headers_here(self.tracer, request, root)
            adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
            resp = self.session.get('mock://datadog/foo')
            assert 200 == resp.status_code
            assert 'bar' == resp.text

        spans = self.tracer.writer.spans
        root, req = spans
        assert 'root' == root.name
        assert 'requests.request' == req.name
        assert root.trace_id == req.trace_id
        assert root.span_id == req.parent_id

    def test_propagation_false(self):
        # ensure distributed tracing can be disabled manually
        cfg = config.get_from(self.session)
        cfg['distributed_tracing'] = False
        adapter = Adapter()
        self.session.mount('mock', adapter)

        with self.tracer.trace('root'):
            def matcher(request):
                return self.headers_not_here(self.tracer, request)
            adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
            resp = self.session.get('mock://datadog/foo')
            assert 200 == resp.status_code
            assert 'bar' == resp.text

    def test_propagation_true_legacy_default(self):
        # [Backward compatibility]: ensure users can switch the distributed
        # tracing flag using the `Session` attribute
        adapter = Adapter()
        self.session.mount('mock', adapter)

        with self.tracer.trace('root') as root:
            def matcher(request):
                return self.headers_here(self.tracer, request, root)
            adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
            resp = self.session.get('mock://datadog/foo')
            assert 200 == resp.status_code
            assert 'bar' == resp.text

        spans = self.tracer.writer.spans
        root, req = spans
        assert 'root' == root.name
        assert 'requests.request' == req.name
        assert root.trace_id == req.trace_id
        assert root.span_id == req.parent_id

    def test_propagation_true_legacy(self):
        # [Backward compatibility]: ensure users can switch the distributed
        # tracing flag using the `Session` attribute
        adapter = Adapter()
        self.session.mount('mock', adapter)
        self.session.distributed_tracing = True

        with self.tracer.trace('root') as root:
            def matcher(request):
                return self.headers_here(self.tracer, request, root)
            adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
            resp = self.session.get('mock://datadog/foo')
            assert 200 == resp.status_code
            assert 'bar' == resp.text

        spans = self.tracer.writer.spans
        root, req = spans
        assert 'root' == root.name
        assert 'requests.request' == req.name
        assert root.trace_id == req.trace_id
        assert root.span_id == req.parent_id

    def test_propagation_false_legacy(self):
        # [Backward compatibility]: ensure users can switch the distributed
        # tracing flag using the `Session` attribute
        adapter = Adapter()
        self.session.mount('mock', adapter)
        self.session.distributed_tracing = False

        with self.tracer.trace('root'):
            def matcher(request):
                return self.headers_not_here(self.tracer, request)
            adapter.register_uri('GET', 'mock://datadog/foo', additional_matcher=matcher, text='bar')
            resp = self.session.get('mock://datadog/foo')
            assert 200 == resp.status_code
            assert 'bar' == resp.text



from copy import deepcopy

from ddtrace.settings import Config, HttpConfig, IntegrationConfig

from tests import BaseTestCase


class TestConfig(BaseTestCase):
    def test_environment_analytics_enabled(self):
        with self.override_env(dict(DD_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED='False')):
            config = Config()
            self.assertFalse(config.analytics_enabled)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED='False')):
            config = Config()
            self.assertFalse(config.analytics_enabled)

    def test_environment_analytics_overrides(self):
        with self.override_env(dict(DD_ANALYTICS_ENABLED='False', DD_TRACE_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED='False', DD_TRACE_ANALYTICS_ENABLED='False')):
            config = Config()
            self.assertFalse(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED='True', DD_TRACE_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED='True', DD_TRACE_ANALYTICS_ENABLED='False')):
            config = Config()
            self.assertFalse(config.analytics_enabled)

    def test_logs_injection(self):
        with self.override_env(dict(DD_LOGS_INJECTION='True')):
            config = Config()
            self.assertTrue(config.logs_injection)

        with self.override_env(dict(DD_LOGS_INJECTION='false')):
            config = Config()
            self.assertFalse(config.logs_injection)

    def test_service(self):
        # If none is provided the default should be ``None``
        with self.override_env(dict()):
            config = Config()
            self.assertEqual(config.service, None)

        with self.override_env(dict(DD_SERVICE="my-service")):
            config = Config()
            self.assertEqual(config.service, "my-service")

    def test_http_config(self):
        config = Config()
        config._add("django", dict())
        assert config.django.trace_query_string is None
        config.http.trace_query_string = True
        assert config.http.trace_query_string is True
        assert config.django.trace_query_string is True

        # Integration usage
        config = Config()
        config._add("django", dict())
        config.django.http.trace_query_string = True
        assert config.http.trace_query_string is None
        assert config.django.trace_query_string is True
        assert config.django.http.trace_query_string is True


class TestHttpConfig(BaseTestCase):

    def test_trace_headers(self):
        http_config = HttpConfig()
        http_config.trace_headers('some_header')
        assert http_config.header_is_traced('some_header')
        assert not http_config.header_is_traced('some_other_header')

    def test_trace_headers_whitelist_case_insensitive(self):
        http_config = HttpConfig()
        http_config.trace_headers('some_header')
        assert http_config.header_is_traced('sOmE_hEaDeR')
        assert not http_config.header_is_traced('some_other_header')

    def test_trace_multiple_headers(self):
        http_config = HttpConfig()
        http_config.trace_headers(['some_header_1', 'some_header_2'])
        assert http_config.header_is_traced('some_header_1')
        assert http_config.header_is_traced('some_header_2')
        assert not http_config.header_is_traced('some_header_3')

    def test_empty_entry_do_not_raise_exception(self):
        http_config = HttpConfig()
        http_config.trace_headers('')

        assert not http_config.header_is_traced('some_header_1')

    def test_none_entry_do_not_raise_exception(self):
        http_config = HttpConfig()
        http_config.trace_headers(None)
        assert not http_config.header_is_traced('some_header_1')

    def test_is_header_tracing_configured(self):
        http_config = HttpConfig()
        assert not http_config.is_header_tracing_configured
        http_config.trace_headers('some_header')
        assert http_config.is_header_tracing_configured

    def test_header_is_traced_case_insensitive(self):
        http_config = HttpConfig()
        http_config.trace_headers('sOmE_hEaDeR')
        assert http_config.header_is_traced('SoMe_HeAdEr')
        assert not http_config.header_is_traced('some_other_header')

    def test_header_is_traced_false_for_empty_header(self):
        http_config = HttpConfig()
        http_config.trace_headers('some_header')
        assert not http_config.header_is_traced('')

    def test_header_is_traced_false_for_none_header(self):
        http_config = HttpConfig()
        http_config.trace_headers('some_header')
        assert not http_config.header_is_traced(None)


class TestIntegrationConfig(BaseTestCase):
    def setUp(self):
        self.config = Config()
        self.integration_config = IntegrationConfig(self.config, 'test')

    def test_is_a_dict(self):
        assert isinstance(self.integration_config, dict)

    def test_allow_item_access(self):
        self.integration_config['setting'] = 'value'

        # Can be accessed both as item and attr accessor
        assert self.integration_config.setting == 'value'
        assert self.integration_config['setting'] == 'value'

    def test_allow_attr_access(self):
        self.integration_config.setting = 'value'

        # Can be accessed both as item and attr accessor
        assert self.integration_config.setting == 'value'
        assert self.integration_config['setting'] == 'value'

    def test_allow_both_access(self):
        self.integration_config.setting = 'value'
        assert self.integration_config['setting'] == 'value'
        assert self.integration_config.setting == 'value'

        self.integration_config['setting'] = 'new-value'
        assert self.integration_config.setting == 'new-value'
        assert self.integration_config['setting'] == 'new-value'

    def test_allow_configuring_http(self):
        self.integration_config.http.trace_headers('integration_header')
        assert self.integration_config.http.header_is_traced('integration_header')
        assert not self.integration_config.http.header_is_traced('other_header')

    def test_allow_exist_both_global_and_integration_config(self):
        self.config.trace_headers('global_header')
        assert self.integration_config.header_is_traced('global_header')

        self.integration_config.http.trace_headers('integration_header')
        assert self.integration_config.header_is_traced('integration_header')
        assert not self.integration_config.header_is_traced('global_header')
        assert not self.config.header_is_traced('integration_header')

    def test_environment_analytics_enabled(self):
        # default
        self.assertFalse(self.config.analytics_enabled)
        self.assertIsNone(self.config.foo.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.analytics_enabled)
            self.assertIsNone(config.foo.analytics_enabled)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.analytics_enabled)
            self.assertIsNone(config.foo.analytics_enabled)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 1.0)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED='True')):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 1.0)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED='False')):
            config = Config()
            self.assertFalse(config.foo.analytics_enabled)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED='False')):
            config = Config()
            self.assertFalse(config.foo.analytics_enabled)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED='True', DD_FOO_ANALYTICS_SAMPLE_RATE='0.5')):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 0.5)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED='True', DD_TRACE_FOO_ANALYTICS_SAMPLE_RATE='0.5')):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 0.5)

    def test_analytics_enabled_attribute(self):
        """" Confirm environment variable and kwargs are handled properly """
        ic = IntegrationConfig(self.config, 'foo', analytics_enabled=True)
        self.assertTrue(ic.analytics_enabled)

        ic = IntegrationConfig(self.config, 'foo', analytics_enabled=False)
        self.assertFalse(ic.analytics_enabled)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED='True')):
            ic = IntegrationConfig(self.config, 'foo', analytics_enabled=False)
            self.assertFalse(ic.analytics_enabled)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED='True')):
            ic = IntegrationConfig(self.config, 'foo', analytics_enabled=False)
            self.assertFalse(ic.analytics_enabled)

    def test_get_analytics_sample_rate(self):
        """" Check method for accessing sample rate based on configuration """
        ic = IntegrationConfig(self.config, 'foo', analytics_enabled=True, analytics_sample_rate=0.5)
        self.assertEqual(ic.get_analytics_sample_rate(), 0.5)

        ic = IntegrationConfig(self.config, 'foo', analytics_enabled=True)
        self.assertEqual(ic.get_analytics_sample_rate(), 1.0)

        ic = IntegrationConfig(self.config, 'foo', analytics_enabled=False)
        self.assertIsNone(ic.get_analytics_sample_rate())

        with self.override_env(dict(DD_ANALYTICS_ENABLED='True')):
            config = Config()
            ic = IntegrationConfig(config, 'foo')
            self.assertEqual(ic.get_analytics_sample_rate(use_global_config=True), 1.0)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED='True')):
            config = Config()
            ic = IntegrationConfig(config, 'foo')
            self.assertEqual(ic.get_analytics_sample_rate(use_global_config=True), 1.0)

        with self.override_env(dict(DD_ANALYTICS_ENABLED='False')):
            config = Config()
            ic = IntegrationConfig(config, 'foo')
            self.assertIsNone(ic.get_analytics_sample_rate(use_global_config=True))

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED='False')):
            config = Config()
            ic = IntegrationConfig(config, 'foo')
            self.assertIsNone(ic.get_analytics_sample_rate(use_global_config=True))

    def test_deepcopy(self):
        ic = IntegrationConfig(
            self.config, 'foo', analytics_enabled=True, analytics_sample_rate=0.5, deep={"field": "sodeep"}
        )
        cpy = deepcopy(ic)
        assert isinstance(cpy, IntegrationConfig)
        assert cpy == ic
        assert cpy.deep["field"] == ic.deep["field"]
        ic.deep["field"] = ""
        assert cpy.deep["field"] != ic.deep["field"]

    def test_copy(self):
        ic = IntegrationConfig(self.config, 'foo', analytics_enabled=True, analytics_sample_rate=0.5)
        copy = ic.copy()
        assert isinstance(copy, IntegrationConfig)

        for key in ic:
            assert key in copy

        for key in copy:
            assert key in ic

        assert ic == copy

    def test_service(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service is None

    @BaseTestCase.run_in_subprocess(env_overrides=dict(DD_FOO_SERVICE="foo-svc"))
    def test_service_env_var(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service == "foo-svc"

    @BaseTestCase.run_in_subprocess(env_overrides=dict(DATADOG_FOO_SERVICE="foo-svc"))
    def test_service_env_var_legacy(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service == "foo-svc"

    @BaseTestCase.run_in_subprocess(env_overrides=dict(DD_FOO_SERVICE_NAME="foo-svc"))
    def test_service_name_env_var(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service == "foo-svc"

    @BaseTestCase.run_in_subprocess(env_overrides=dict(DATADOG_FOO_SERVICE_NAME="foo-svc"))
    def test_service_name_env_var_legacy(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service == "foo-svc"



# -*- coding: utf-8 -*-
import unittest

import flask

from ddtrace.vendor import wrapt
from ddtrace.ext import http
from ddtrace import Pin

from tests.tracer.test_tracer import get_dummy_tracer
from ... import assert_is_measured, assert_span_http_status_code


class FlaskAutopatchTestCase(unittest.TestCase):
    def setUp(self):
        self.tracer = get_dummy_tracer()
        self.app = flask.Flask(__name__)
        Pin.override(self.app, service='test-flask', tracer=self.tracer)
        self.client = self.app.test_client()

    def test_patched(self):
        """
        When using ddtrace-run
            Then the `flask` module is patched
        """
        # DEV: We have great test coverage in tests.contrib.flask,
        #      we only need basic tests here to assert `ddtrace-run` patched thingsa

        # Assert module is marked as patched
        self.assertTrue(flask._datadog_patch)

        # Assert our instance of flask.app.Flask is patched
        self.assertTrue(isinstance(self.app.add_url_rule, wrapt.ObjectProxy))
        self.assertTrue(isinstance(self.app.wsgi_app, wrapt.ObjectProxy))

        # Assert the base module flask.app.Flask methods are patched
        self.assertTrue(isinstance(flask.app.Flask.add_url_rule, wrapt.ObjectProxy))
        self.assertTrue(isinstance(flask.app.Flask.wsgi_app, wrapt.ObjectProxy))

    def test_request(self):
        """
        When using ddtrace-run
            When making a request to flask app
                We generate the expected spans
        """
        @self.app.route('/')
        def index():
            return 'Hello Flask', 200

        res = self.client.get('/')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.data, b'Hello Flask')

        spans = self.tracer.writer.pop()
        self.assertEqual(len(spans), 8)

        self.assertListEqual(
            [
                'flask.request',
                'flask.try_trigger_before_first_request_functions',
                'flask.preprocess_request',
                'flask.dispatch_request',
                'tests.contrib.flask_autopatch.test_flask_autopatch.index',
                'flask.process_response',
                'flask.do_teardown_request',
                'flask.do_teardown_appcontext',
            ],
            [s.name for s in spans],
        )

        # Assert span services
        for span in spans:
            self.assertEqual(span.service, 'test-flask')

        # Root request span
        req_span = spans[0]
        assert_is_measured(req_span)
        self.assertEqual(req_span.service, 'test-flask')
        self.assertEqual(req_span.name, 'flask.request')
        self.assertEqual(req_span.resource, 'GET /')
        self.assertEqual(req_span.span_type, 'web')
        self.assertEqual(req_span.error, 0)
        self.assertIsNone(req_span.parent_id)

        # Request tags
        for tag in ['flask.version', 'http.url', 'http.method', 'http.status_code',
                    'flask.endpoint', 'flask.url_rule']:
            assert tag in req_span.meta

        self.assertEqual(req_span.get_tag('flask.endpoint'), 'index')
        self.assertEqual(req_span.get_tag('flask.url_rule'), '/')
        self.assertEqual(req_span.get_tag('http.method'), 'GET')
        self.assertEqual(req_span.get_tag(http.URL), 'http://localhost/')
        assert_span_http_status_code(req_span, 200)

        # Handler span
        handler_span = spans[4]
        self.assertEqual(handler_span.service, 'test-flask')
        self.assertEqual(handler_span.name, 'tests.contrib.flask_autopatch.test_flask_autopatch.index')
        self.assertEqual(handler_span.resource, '/')
        self.assertEqual(req_span.error, 0)



import os
import sys

# DEV: We must append to sys path before importing ddtrace_run
sys.path.append('.')
from ddtrace.commands import ddtrace_run  # noqa

os.environ['PYTHONPATH'] = '{}:{}'.format(os.getenv('PYTHONPATH'), os.path.abspath('.'))
ddtrace_run.main()



from .collector import ValueCollector
from .constants import (
    SERVICE,
    LANG_INTERPRETER,
    LANG_VERSION,
    LANG,
    TRACER_VERSION,
)
from ...constants import ENV_KEY


class RuntimeTagCollector(ValueCollector):
    periodic = False
    value = []


class TracerTagCollector(RuntimeTagCollector):
    """Tag collector for the ddtrace Tracer"""

    required_modules = ["ddtrace"]

    def collect_fn(self, keys):
        ddtrace = self.modules.get("ddtrace")
        tags = [(SERVICE, service) for service in ddtrace.tracer._services]
        if ENV_KEY in ddtrace.tracer.tags:
            tags.append((ENV_KEY, ddtrace.tracer.tags[ENV_KEY]))
        return tags


class PlatformTagCollector(RuntimeTagCollector):
    """Tag collector for the Python interpreter implementation.

    Tags collected:
    - ``lang_interpreter``:

      * For CPython this is 'CPython'.
      * For Pypy this is ``PyPy``
      * For Jython this is ``Jython``

    - `lang_version``,  eg ``2.7.10``
    - ``lang`` e.g. ``Python``
    - ``tracer_version`` e.g. ``0.29.0``

    """

    required_modules = ("platform", "ddtrace")

    def collect_fn(self, keys):
        platform = self.modules.get("platform")
        ddtrace = self.modules.get("ddtrace")
        tags = [
            (LANG, "python"),
            (LANG_INTERPRETER, platform.python_implementation()),
            (LANG_VERSION, platform.python_version()),
            (TRACER_VERSION, ddtrace.__version__),
        ]
        return tags



"""Test compatibility with old ddtrace.profile name."""
import subprocess
import os


def test_call_script(monkeypatch):
    # Set a very short timeout to exit fast
    monkeypatch.setenv("DD_PROFILING_API_TIMEOUT", "0.1")
    subp = subprocess.Popen(
        ["python", os.path.join(os.path.dirname(__file__), "compat_program.py")], stdout=subprocess.PIPE
    )
    stdout, stderr = subp.communicate()
    assert subp.wait() == 42
    hello, interval, stacks = stdout.decode().strip().split("\n")
    assert hello == "hello world"
    assert float(interval) >= 0.01
    assert int(stacks) >= 1



import aiobotocore.session

from ddtrace import Pin
from contextlib import contextmanager


LOCALSTACK_ENDPOINT_URL = {
    's3': 'http://localhost:5000',
    'ec2': 'http://localhost:5001',
    'kms': 'http://localhost:5002',
    'sqs': 'http://localhost:5003',
    'lambda': 'http://localhost:5004',
    'kinesis': 'http://localhost:5005',
}


@contextmanager
def aiobotocore_client(service, tracer):
    """Helper function that creates a new aiobotocore client so that
    it is closed at the end of the context manager.
    """
    session = aiobotocore.session.get_session()
    endpoint = LOCALSTACK_ENDPOINT_URL[service]
    client = session.create_client(
        service,
        region_name='us-west-2',
        endpoint_url=endpoint,
        aws_access_key_id='aws',
        aws_secret_access_key='aws',
        aws_session_token='aws',
    )
    Pin.override(client, tracer=tracer)
    try:
        yield client
    finally:
        client.close()



import sys

from ddtrace import tracer as global_tracer
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.ext import SpanTypes, http
from ddtrace.http import store_request_headers, store_response_headers
from ddtrace.propagation.http import HTTPPropagator
from ddtrace.settings import config

from ...compat import reraise
from ...internal.logger import get_logger
from .utils import guarantee_single_callable

log = get_logger(__name__)

config._add("asgi", dict(service_name=config._get_service(default="asgi"), distributed_tracing=True))

ASGI_VERSION = "asgi.version"
ASGI_SPEC_VERSION = "asgi.spec_version"


def bytes_to_str(str_or_bytes):
    return str_or_bytes.decode() if isinstance(str_or_bytes, bytes) else str_or_bytes


def _extract_tags_from_scope(scope):
    tags = {}

    http_method = scope.get("method")
    if http_method:
        tags[http.METHOD] = http_method

    if config.asgi.trace_query_string:
        query_string = scope.get("query_string")
        if len(query_string) > 0:
            tags[http.QUERY_STRING] = bytes_to_str(query_string)
    else:
        query_string = None

    server = scope.get("server")
    if server and len(server) == 2:
        port = server[1]
        server_host = server[0] + (":" + str(port) if port is not None and port != 80 else "")
        full_path = scope.get("root_path", "") + scope.get("path", "")
        http_url = scope.get("scheme", "http") + "://" + server_host + full_path
        tags[http.URL] = http_url

    http_version = scope.get("http_version")
    if http_version:
        tags[http.VERSION] = http_version

    scope_asgi = scope.get("asgi")

    if scope_asgi and "version" in scope_asgi:
        tags[ASGI_VERSION] = scope_asgi["version"]

    if scope_asgi and "spec_version" in scope_asgi:
        tags[ASGI_SPEC_VERSION] = scope_asgi["spec_version"]

    return tags


def _extract_headers(scope):
    headers = scope.get("headers")
    if headers:
        # headers: (Iterable[[byte string, byte string]])
        return dict((bytes_to_str(k), bytes_to_str(v)) for (k, v) in headers)
    return {}


def _default_handle_exception_span(exc, span):
    """Default handler for exception for span"""
    span.set_tag(http.STATUS_CODE, 500)


class TraceMiddleware:
    """
    ASGI application middleware that traces the requests.

    Args:
        app: The ASGI application.
        tracer: Custom tracer. Defaults to the global tracer.
    """

    def __init__(self, app, tracer=global_tracer, handle_exception_span=_default_handle_exception_span):
        self.app = guarantee_single_callable(app)
        self.tracer = tracer
        self.handle_exception_span = handle_exception_span

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        headers = _extract_headers(scope)

        if config.asgi.distributed_tracing:
            propagator = HTTPPropagator()
            context = propagator.extract(headers)
            if context.trace_id:
                self.tracer.context_provider.activate(context)

        resource = "{} {}".format(scope["method"], scope["path"])

        span = self.tracer.trace(
            name="asgi.request",
            service=config.asgi.service_name,
            resource=resource,
            span_type=SpanTypes.HTTP,
        )

        sample_rate = config.asgi.get_analytics_sample_rate(use_global_config=True)
        if sample_rate is not None:
            span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, sample_rate)

        tags = _extract_tags_from_scope(scope)
        span.set_tags(tags)

        store_request_headers(headers, span, config.asgi)

        async def wrapped_send(message):
            if span and message.get("type") == "http.response.start":
                if "status" in message:
                    status_code = message["status"]
                    span.set_tag(http.STATUS_CODE, status_code)
                    if 500 <= int(status_code) < 600:
                        span.error = 1

                if "headers" in message:
                    store_response_headers(message["headers"], span, config.asgi)

            return await send(message)

        try:
            return await self.app(scope, receive, wrapped_send)
        except Exception as exc:
            (exc_type, exc_val, exc_tb) = sys.exc_info()
            span.set_exc_info(exc_type, exc_val, exc_tb)
            self.handle_exception_span(exc, span)
            reraise(exc_type, exc_val, exc_tb)
        finally:
            span.finish()



import contextlib
import mock
import sys

import pytest

from ddtrace.internal import import_hooks

from tests.subprocesstest import run_in_subprocess, SubprocessTestCase


import_hooks.patch()


@pytest.fixture
def hooks():
    import_hooks.hooks.reset()
    try:
        yield import_hooks.hooks
    finally:
        import_hooks.hooks.reset()


@contextlib.contextmanager
def remove_module(name):
    was_loaded = name in sys.modules
    try:
        # Ensure urllib is not loaded
        if was_loaded:
            del sys.modules[name]
        yield
    finally:
        if not was_loaded:
            del sys.modules[name]


def test_import_hooks(hooks):
    """
    When registering an import hook
        Gets called after the module was imported
    """
    # Ensure our module is not yet loaded
    mod_name = "urllib"
    with remove_module(mod_name):
        # Register our hook (when the module is not loaded)
        module_hook = mock.Mock()
        import_hooks.register_module_hook(mod_name, module_hook)

        # Import the module being hooked
        import urllib

        # Ensure we called our hook with the module
        # DEV: Slightly redundant to check twice, but good to be sure
        module_hook.assert_called_once_with(urllib)
        module_hook.assert_called_once_with(sys.modules[mod_name])


@run_in_subprocess
class ImportHookTestCase(SubprocessTestCase):
    def test_register_then_import(self):
        """
        When an import hook is registered before importing a module
            The import hook should run when the module is imported
        """
        module_hook = mock.Mock()
        import_hooks.register_module_hook("tests.test_module", module_hook)

        # Hook should not be called after register
        module_hook.assert_not_called()

        import tests.test_module

        # Module imported so the hook should be called
        module_hook.assert_called_once_with(tests.test_module)

    def test_import_then_register(self):
        """
        When a module is imported before the import hook is registered
            The import hook should run immediately when registered
        """
        module_hook = mock.Mock()
        import tests.test_module

        module_hook.assert_not_called()
        import_hooks.register_module_hook("tests.test_module", module_hook)

        # Hook should be called on register
        module_hook.assert_called_once_with(tests.test_module)

    def test_no_double_call(self):
        """
        When a module is imported multiple times
            The import hook should only be run once
        """
        module_hook = mock.Mock()
        import_hooks.register_module_hook("tests.test_module", module_hook)
        import tests.test_module
        import tests.test_module  # noqa
        import tests.test_module  # noqa
        import tests.test_module  # noqa

        # Hook should be called only once
        module_hook.assert_called_once_with(tests.test_module)

    def test_register_deregister(self):
        """
        When an import hook is registered and subsequently deregistered
            The import hook should not be called
        """
        module_hook = mock.Mock()
        import_hooks.register_module_hook("tests.test_module", module_hook)
        import_hooks.hooks.deregister("tests.test_module", module_hook)
        import tests.test_module  # noqa

        # Hook should not be called
        module_hook.assert_not_called()

    def test_register_multiple_modules(self):
        """
        When registering module hooks on multiple modules
            Each hook is called when their respective module is loaded
        """
        test_module_hook = mock.Mock()
        test_module_hook2 = mock.Mock()
        test_module2_hook = mock.Mock()

        import_hooks.register_module_hook("tests.test_module", test_module_hook)
        import_hooks.register_module_hook("tests.test_module", test_module_hook2)
        import_hooks.register_module_hook("tests.test_module2", test_module2_hook)

        test_module_hook.assert_not_called()
        test_module_hook2.assert_not_called()
        test_module2_hook.assert_not_called()

        import tests.test_module2

        test_module_hook.assert_not_called()
        test_module_hook2.assert_not_called()
        test_module2_hook.assert_called_once_with(tests.test_module2)

        import tests.test_module

        test_module_hook.assert_called_once_with(tests.test_module)
        test_module_hook2.assert_called_once_with(tests.test_module)
        test_module2_hook.assert_called_once_with(tests.test_module2)



# -*- coding: utf-8 -*-
import os

from flask import Flask, render_template


class TestError(Exception):
    pass


class HandleMe(Exception):
    pass


def create_app():
    """Initializes a new Flask application. This method is required to
    be sure each time a test is executed, the Flask app is always new
    and without any tracing side effect from the previous execution.
    """
    cur_dir = os.path.dirname(os.path.realpath(__file__))
    tmpl_path = os.path.join(cur_dir, 'test_templates')
    app = Flask(__name__, template_folder=tmpl_path)

    @app.route('/')
    def index():
        return 'hello'

    @app.route('/error')
    def error():
        raise TestError()

    @app.route('/handleme')
    def handle_me():
        raise HandleMe()

    @app.route('/fatal')
    def fatal():
        1 / 0

    @app.route('/tmpl')
    def tmpl():
        return render_template('test.html', world='earth')

    @app.route('/tmpl/err')
    def tmpl_err():
        return render_template('err.html')

    @app.route('/tmpl/render_err')
    def tmpl_render_err():
        return render_template('render_err.html')

    @app.route('/child')
    def child():
        with app._tracer.trace('child') as span:
            span.set_tag('a', 'b')
            return 'child'

    @app.route('/custom_span')
    def custom_span():
        span = app._tracer.current_span()
        assert span
        span.resource = 'overridden'
        return 'hiya'

    def unicode_view():
        return u'üŋïĉóđē'

    app.add_url_rule(
        u'/üŋïĉóđē',
        u'üŋïĉóđē',
        unicode_view,
    )

    @app.errorhandler(TestError)
    def handle_my_exception(e):
        assert isinstance(e, TestError)
        return 'error', 500

    @app.errorhandler(HandleMe)
    def err_to_202(e):
        assert isinstance(e, HandleMe)
        return 'handled', 202

    return app



from ddtrace.internal._queue import TraceQueue


def test_queue_no_limit():
    q = TraceQueue()
    for i in range(0, 10000):
        q.put([i])

    items = q.get()
    assert len(items) == 10000
    for i in range(0, 10000):
        assert items[i] == [i]

    dropped, accepted, lengths = q.pop_stats()
    assert dropped == 0
    assert accepted == 10000
    assert lengths == 10000


def test_queue_full():
    q = TraceQueue(maxsize=3)
    q.put([1])
    q.put(2)
    q.put([3])
    q.put([4, 4])
    assert len(q) == 3

    state = list(q._queue)
    assert state == [[1], 2, [4, 4]] or state == [[1], [4, 4], [3]] or state == [[4, 4], 2, [3]]
    assert q._dropped == 1
    assert q._accepted == 4
    assert q._accepted_lengths == 5

    dropped, accepted, accepted_lengths = q.pop_stats()
    assert dropped == 1
    assert accepted == 4
    assert accepted_lengths == 5


def test_queue_get():
    q = TraceQueue(maxsize=3)
    q.put(1)
    q.put(2)
    assert q.get() == [1, 2]
    assert q.get() == []


def test_multiple_queues():
    q1 = TraceQueue()
    q2 = TraceQueue()

    q1.put(1)
    q2.put(1)

    assert len(q1) == 1
    assert len(q2) == 1

    q1.get()
    assert len(q1) == 0
    assert len(q2) == 1

    q2.get()
    assert len(q1) == 0
    assert len(q2) == 0


def test_queue_overflow():
    q = TraceQueue(maxsize=1000)

    for i in range(10000):
        q.put([])
        assert len(q) <= 1000

    dropped, accepted, accepted_lengths = q.pop_stats()
    assert dropped == 9000
    assert accepted == 10000
    assert accepted_lengths == 0



import mock
import time

import pytest
from unittest.case import SkipTest

from ddtrace.context import Context
from ddtrace.constants import (
    ANALYTICS_SAMPLE_RATE_KEY, VERSION_KEY,
    SERVICE_VERSION_KEY, SPAN_MEASURED_KEY, ENV_KEY,
)
from ddtrace.span import Span
from ddtrace.ext import SpanTypes, errors, priority
from tests import TracerTestCase, assert_is_measured, assert_is_not_measured


class SpanTestCase(TracerTestCase):
    def test_ids(self):
        s = Span(tracer=None, name='span.test')
        assert s.trace_id
        assert s.span_id
        assert not s.parent_id

        s2 = Span(tracer=None, name='t', trace_id=1, span_id=2, parent_id=1)
        assert s2.trace_id == 1
        assert s2.span_id == 2
        assert s2.parent_id == 1

    def test_tags(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag('a', 'a')
        s.set_tag('b', 1)
        s.set_tag('c', '1')
        d = s.to_dict()
        assert d['meta'] == dict(a='a', c='1')
        assert d['metrics'] == dict(b=1)

    def test_numeric_tags(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag('negative', -1)
        s.set_tag('zero', 0)
        s.set_tag('positive', 1)
        s.set_tag('large_int', 2**53)
        s.set_tag('really_large_int', (2**53) + 1)
        s.set_tag('large_negative_int', -(2**53))
        s.set_tag('really_large_negative_int', -((2**53) + 1))
        s.set_tag('float', 12.3456789)
        s.set_tag('negative_float', -12.3456789)
        s.set_tag('large_float', 2.0**53)
        s.set_tag('really_large_float', (2.0**53) + 1)

        d = s.to_dict()
        assert d['meta'] == dict(
            really_large_int=str(((2**53) + 1)),
            really_large_negative_int=str(-((2**53) + 1)),
        )
        assert d['metrics'] == {
            'negative': -1,
            'zero': 0,
            'positive': 1,
            'large_int': 2**53,
            'large_negative_int': -(2**53),
            'float': 12.3456789,
            'negative_float': -12.3456789,
            'large_float': 2.0**53,
            'really_large_float': (2.0**53) + 1,
        }

    def test_set_tag_bool(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag('true', True)
        s.set_tag('false', False)

        d = s.to_dict()
        assert d['meta'] == dict(true='True', false='False')
        assert 'metrics' not in d

    def test_set_tag_metric(self):
        s = Span(tracer=None, name='test.span')

        s.set_tag('test', 'value')
        assert s.meta == dict(test='value')
        assert s.metrics == dict()

        s.set_tag('test', 1)
        assert s.meta == dict()
        assert s.metrics == dict(test=1)

    def test_set_valid_metrics(self):
        s = Span(tracer=None, name='test.span')
        s.set_metric('a', 0)
        s.set_metric('b', -12)
        s.set_metric('c', 12.134)
        s.set_metric('d', 1231543543265475686787869123)
        s.set_metric('e', '12.34')
        d = s.to_dict()
        expected = {
            'a': 0,
            'b': -12,
            'c': 12.134,
            'd': 1231543543265475686787869123,
            'e': 12.34,
        }
        assert d['metrics'] == expected

    def test_set_invalid_metric(self):
        s = Span(tracer=None, name='test.span')

        invalid_metrics = [
            None,
            {},
            [],
            s,
            'quarante-douze',
            float('nan'),
            float('inf'),
            1j
        ]

        for i, m in enumerate(invalid_metrics):
            k = str(i)
            s.set_metric(k, m)
            assert s.get_metric(k) is None

    def test_set_numpy_metric(self):
        try:
            import numpy as np
        except ImportError:
            raise SkipTest('numpy not installed')
        s = Span(tracer=None, name='test.span')
        s.set_metric('a', np.int64(1))
        assert s.get_metric('a') == 1
        assert type(s.get_metric('a')) == float

    def test_tags_not_string(self):
        # ensure we can cast as strings
        class Foo(object):
            def __repr__(self):
                1 / 0

        s = Span(tracer=None, name='test.span')
        s.set_tag('a', Foo())

    def test_finish(self):
        # ensure finish will record a span
        ctx = Context()
        s = Span(self.tracer, 'test.span', context=ctx)
        ctx.add_span(s)
        assert s.duration is None

        sleep = 0.05
        with s as s1:
            assert s is s1
            time.sleep(sleep)
        assert s.duration >= sleep, '%s < %s' % (s.duration, sleep)
        self.assert_span_count(1)

    def test_finish_no_tracer(self):
        # ensure finish works with no tracer without raising exceptions
        s = Span(tracer=None, name='test.span')
        s.finish()

    def test_finish_called_multiple_times(self):
        # we should only record a span the first time finish is called on it
        ctx = Context()
        s = Span(self.tracer, 'bar', context=ctx)
        ctx.add_span(s)
        s.finish()
        s.finish()
        self.assert_span_count(1)

    def test_finish_set_span_duration(self):
        # If set the duration on a span, the span should be recorded with this
        # duration
        s = Span(tracer=None, name='test.span')
        s.duration = 1337.0
        s.finish()
        assert s.duration == 1337.0

    def test_traceback_with_error(self):
        s = Span(None, 'test.span')
        try:
            1 / 0
        except ZeroDivisionError:
            s.set_traceback()
        else:
            assert 0, 'should have failed'

        assert s.error
        assert 'by zero' in s.get_tag(errors.ERROR_MSG)
        assert 'ZeroDivisionError' in s.get_tag(errors.ERROR_TYPE)

    def test_traceback_without_error(self):
        s = Span(None, 'test.span')
        s.set_traceback()
        assert not s.error
        assert not s.get_tag(errors.ERROR_MSG)
        assert not s.get_tag(errors.ERROR_TYPE)
        assert 'in test_traceback_without_error' in s.get_tag(errors.ERROR_STACK)

    def test_ctx_mgr(self):
        s = Span(self.tracer, 'bar')
        assert not s.duration
        assert not s.error

        e = Exception('boo')
        try:
            with s:
                time.sleep(0.01)
                raise e
        except Exception as out:
            assert out == e
            assert s.duration > 0, s.duration
            assert s.error
            assert s.get_tag(errors.ERROR_MSG) == 'boo'
            assert 'Exception' in s.get_tag(errors.ERROR_TYPE)
            assert s.get_tag(errors.ERROR_STACK)

        else:
            assert 0, 'should have failed'

    def test_span_type(self):
        s = Span(tracer=None, name='test.span', service='s', resource='r', span_type=SpanTypes.WEB)
        s.set_tag('a', '1')
        s.set_meta('b', '2')
        s.finish()

        d = s.to_dict()
        assert d
        assert d['span_id'] == s.span_id
        assert d['trace_id'] == s.trace_id
        assert d['parent_id'] == s.parent_id
        assert d['meta'] == {'a': '1', 'b': '2'}
        assert d['type'] == 'web'
        assert d['error'] == 0
        assert type(d['error']) == int

    def test_span_to_dict(self):
        s = Span(tracer=None, name='test.span', service='s', resource='r')
        s.span_type = 'foo'
        s.set_tag('a', '1')
        s.set_meta('b', '2')
        s.finish()

        d = s.to_dict()
        assert d
        assert d['span_id'] == s.span_id
        assert d['trace_id'] == s.trace_id
        assert d['parent_id'] == s.parent_id
        assert d['meta'] == {'a': '1', 'b': '2'}
        assert d['type'] == 'foo'
        assert d['error'] == 0
        assert type(d['error']) == int

    def test_span_to_dict_sub(self):
        parent = Span(tracer=None, name='test.span', service='s', resource='r')
        s = Span(tracer=None, name='test.span', service='s', resource='r')
        s._parent = parent
        s.span_type = 'foo'
        s.set_tag('a', '1')
        s.set_meta('b', '2')
        s.finish()

        d = s.to_dict()
        assert d
        assert d['span_id'] == s.span_id
        assert d['trace_id'] == s.trace_id
        assert d['parent_id'] == s.parent_id
        assert d['meta'] == {'a': '1', 'b': '2'}
        assert d['type'] == 'foo'
        assert d['error'] == 0
        assert type(d['error']) == int

    def test_span_boolean_err(self):
        s = Span(tracer=None, name='foo.bar', service='s', resource='r')
        s.error = True
        s.finish()

        d = s.to_dict()
        assert d
        assert d['error'] == 1
        assert type(d['error']) == int

    @mock.patch('ddtrace.span.log')
    def test_numeric_tags_none(self, span_log):
        s = Span(tracer=None, name='test.span')
        s.set_tag(ANALYTICS_SAMPLE_RATE_KEY, None)
        d = s.to_dict()
        assert d
        assert 'metrics' not in d

        # Ensure we log a debug message
        span_log.debug.assert_called_once_with(
            'ignoring not number metric %s:%s',
            ANALYTICS_SAMPLE_RATE_KEY,
            None,
        )

    def test_numeric_tags_true(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag(ANALYTICS_SAMPLE_RATE_KEY, True)
        d = s.to_dict()
        assert d
        expected = {
            ANALYTICS_SAMPLE_RATE_KEY: 1.0
        }
        assert d['metrics'] == expected

    def test_numeric_tags_value(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag(ANALYTICS_SAMPLE_RATE_KEY, 0.5)
        d = s.to_dict()
        assert d
        expected = {
            ANALYTICS_SAMPLE_RATE_KEY: 0.5
        }
        assert d['metrics'] == expected

    def test_numeric_tags_bad_value(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag(ANALYTICS_SAMPLE_RATE_KEY, 'Hello')
        d = s.to_dict()
        assert d
        assert 'metrics' not in d

    def test_set_tag_manual_keep(self):
        ctx = Context()
        s = Span(tracer=None, name='root.span', service='s', resource='r', context=ctx)

        assert s.context == ctx
        assert ctx.sampling_priority != priority.USER_KEEP
        assert s.context.sampling_priority != priority.USER_KEEP
        assert s.meta == dict()

        s.set_tag('manual.keep')
        assert ctx.sampling_priority == priority.USER_KEEP
        assert s.context.sampling_priority == priority.USER_KEEP
        assert s.meta == dict()

        ctx.sampling_priority = priority.AUTO_REJECT
        assert ctx.sampling_priority == priority.AUTO_REJECT
        assert s.context.sampling_priority == priority.AUTO_REJECT
        assert s.meta == dict()

        s.set_tag('manual.keep')
        assert ctx.sampling_priority == priority.USER_KEEP
        assert s.context.sampling_priority == priority.USER_KEEP
        assert s.meta == dict()

    def test_set_tag_manual_drop(self):
        ctx = Context()
        s = Span(tracer=None, name='root.span', service='s', resource='r', context=ctx)

        assert s.context == ctx
        assert ctx.sampling_priority != priority.USER_REJECT
        assert s.context.sampling_priority != priority.USER_REJECT
        assert s.meta == dict()

        s.set_tag('manual.drop')
        assert ctx.sampling_priority == priority.USER_REJECT
        assert s.context.sampling_priority == priority.USER_REJECT
        assert s.meta == dict()

        ctx.sampling_priority = priority.AUTO_REJECT
        assert ctx.sampling_priority == priority.AUTO_REJECT
        assert s.context.sampling_priority == priority.AUTO_REJECT
        assert s.meta == dict()

        s.set_tag('manual.drop')
        assert ctx.sampling_priority == priority.USER_REJECT
        assert s.context.sampling_priority == priority.USER_REJECT
        assert s.meta == dict()

    def test_set_tag_none(self):
        s = Span(tracer=None, name='root.span', service='s', resource='r')
        assert s.meta == dict()

        s.set_tag('custom.key', '100')

        assert s.meta == {'custom.key': '100'}

        s.set_tag('custom.key', None)

        assert s.meta == {'custom.key': 'None'}

    def test_duration_zero(self):
        s = Span(tracer=None, name='foo.bar', service='s', resource='r', start=123)
        s.finish(finish_time=123)
        assert s.duration_ns == 0
        assert s.duration == 0

    def test_start_int(self):
        s = Span(tracer=None, name='foo.bar', service='s', resource='r', start=123)
        assert s.start == 123
        assert s.start_ns == 123000000000

        s = Span(tracer=None, name='foo.bar', service='s', resource='r', start=123.123)
        assert s.start == 123.123
        assert s.start_ns == 123123000000

        s = Span(tracer=None, name='foo.bar', service='s', resource='r', start=123.123)
        s.start = 234567890.0
        assert s.start == 234567890
        assert s.start_ns == 234567890000000000

    def test_duration_int(self):
        s = Span(tracer=None, name='foo.bar', service='s', resource='r')
        s.finish()
        assert isinstance(s.duration_ns, int)
        assert isinstance(s.duration, float)

        s = Span(tracer=None, name='foo.bar', service='s', resource='r', start=123)
        s.finish(finish_time=123.2)
        assert s.duration_ns == 200000000
        assert s.duration == 0.2

        s = Span(tracer=None, name='foo.bar', service='s', resource='r', start=123.1)
        s.finish(finish_time=123.2)
        assert s.duration_ns == 100000000
        assert s.duration == 0.1

        s = Span(tracer=None, name='foo.bar', service='s', resource='r', start=122)
        s.finish(finish_time=123)
        assert s.duration_ns == 1000000000
        assert s.duration == 1

    def test_set_tag_version(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag(VERSION_KEY, '1.2.3')
        assert s.get_tag(VERSION_KEY) == '1.2.3'
        assert s.get_tag(SERVICE_VERSION_KEY) is None

        s.set_tag(SERVICE_VERSION_KEY, 'service.version')
        assert s.get_tag(VERSION_KEY) == 'service.version'
        assert s.get_tag(SERVICE_VERSION_KEY) == 'service.version'

    def test_set_tag_env(self):
        s = Span(tracer=None, name='test.span')
        s.set_tag(ENV_KEY, 'prod')
        assert s.get_tag(ENV_KEY) == 'prod'


@pytest.mark.parametrize(
    "value,assertion",
    [
        (None, assert_is_measured),
        (1, assert_is_measured),
        (1.0, assert_is_measured),
        (-1, assert_is_measured),
        (True, assert_is_measured),
        ("true", assert_is_measured),

        # DEV: Ends up being measured because we do `bool("false")` which is `True`
        ("false", assert_is_measured),

        (0, assert_is_not_measured),
        (0.0, assert_is_not_measured),
        (False, assert_is_not_measured),
    ],
)
def test_set_tag_measured(value, assertion):
    s = Span(tracer=None, name="test.span")
    s.set_tag(SPAN_MEASURED_KEY, value)
    assertion(s)


def test_set_tag_measured_not_set():
    # Span is not measured by default
    s = Span(tracer=None, name="test.span")
    assert_is_not_measured(s)


def test_set_tag_measured_no_value():
    s = Span(tracer=None, name="test.span")
    s.set_tag(SPAN_MEASURED_KEY)
    assert_is_measured(s)


def test_set_tag_measured_change_value():
    s = Span(tracer=None, name="test.span")
    s.set_tag(SPAN_MEASURED_KEY, True)
    assert_is_measured(s)

    s.set_tag(SPAN_MEASURED_KEY, False)
    assert_is_not_measured(s)

    s.set_tag(SPAN_MEASURED_KEY)
    assert_is_measured(s)


@mock.patch('ddtrace.span.log')
def test_span_key(span_log):
    # Span tag keys must be strings
    s = Span(tracer=None, name="test.span")

    s.set_tag(123, True)
    span_log.warning.assert_called_once_with("Ignoring tag pair %s:%s. Key must be a string.", 123, True)
    assert s.get_tag(123) is None
    assert s.get_tag("123") is None

    span_log.reset_mock()

    s.set_tag(None, "val")
    span_log.warning.assert_called_once_with("Ignoring tag pair %s:%s. Key must be a string.", None, "val")
    assert s.get_tag(123.32) is None


def test_span_finished():
    span = Span(None, None)
    assert span.finished is False
    assert span.duration_ns is None

    span.finished = True
    assert span.finished is True
    assert span.duration_ns is not None
    duration = span.duration_ns

    span.finished = True
    assert span.finished is True
    assert span.duration_ns == duration

    span.finished = False
    assert span.finished is False

    span.finished = True
    assert span.finished is True
    assert span.duration_ns != duration



import contextlib
import logging
import mock
import threading

from .. import BaseTestCase
from .test_tracer import get_dummy_tracer

import pytest

from ddtrace.span import Span
from ddtrace.context import Context
from ddtrace.constants import HOSTNAME_KEY
from ddtrace.ext.priority import USER_REJECT, AUTO_REJECT, AUTO_KEEP, USER_KEEP


@pytest.fixture
def tracer_with_debug_logging():
    # All the tracers, dummy or not, shares the same logging object.
    tracer = get_dummy_tracer()
    level = tracer.log.level
    tracer.log.setLevel(logging.DEBUG)
    try:
        yield tracer
    finally:
        tracer.log.setLevel(level)


@mock.patch('logging.Logger.debug')
def test_log_unfinished_spans(log, tracer_with_debug_logging):
    # when the root parent is finished, notify if there are spans still pending
    tracer = tracer_with_debug_logging
    ctx = Context()
    # manually create a root-child trace
    root = Span(tracer=tracer, name='root')
    child_1 = Span(tracer=tracer, name='child_1', trace_id=root.trace_id, parent_id=root.span_id)
    child_2 = Span(tracer=tracer, name='child_2', trace_id=root.trace_id, parent_id=root.span_id)
    child_1._parent = root
    child_2._parent = root
    ctx.add_span(root)
    ctx.add_span(child_1)
    ctx.add_span(child_2)
    # close only the parent
    root.finish()
    unfinished_spans_log = log.call_args_list[-3][0][2]
    child_1_log = log.call_args_list[-2][0][1]
    child_2_log = log.call_args_list[-1][0][1]
    assert 2 == unfinished_spans_log
    assert 'name child_1' in child_1_log
    assert 'name child_2' in child_2_log
    assert 'duration 0.000000s' in child_1_log
    assert 'duration 0.000000s' in child_2_log


class TestTracingContext(BaseTestCase):
    """
    Tests related to the ``Context`` class that hosts the trace for the
    current execution flow.
    """
    @contextlib.contextmanager
    def override_partial_flush(self, ctx, enabled, min_spans):
        original_enabled = ctx._partial_flush_enabled
        original_min_spans = ctx._partial_flush_min_spans

        ctx._partial_flush_enabled = enabled
        ctx._partial_flush_min_spans = min_spans

        try:
            yield
        finally:
            ctx._partial_flush_enabled = original_enabled
            ctx._partial_flush_min_spans = original_min_spans

    def test_add_span(self):
        # it should add multiple spans
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        assert 1 == len(ctx._trace)
        assert 'fake_span' == ctx._trace[0].name
        assert ctx == span.context

    def test_context_sampled(self):
        # a context is sampled if the spans are sampled
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        span.finish()
        trace, sampled = ctx.get()
        assert sampled is True
        assert ctx.sampling_priority is None

    def test_context_priority(self):
        # a context is sampled if the spans are sampled
        ctx = Context()
        for priority in [USER_REJECT, AUTO_REJECT, AUTO_KEEP, USER_KEEP, None, 999]:
            ctx.sampling_priority = priority
            span = Span(tracer=None, name=('fake_span_%s' % repr(priority)))
            ctx.add_span(span)
            span.finish()
            # It's "normal" to have sampled be true even when priority sampling is
            # set to 0 or -1. It would stay false even even with priority set to 2.
            # The only criteria to send (or not) the spans to the agent should be
            # this "sampled" attribute, as it's tightly related to the trace weight.
            assert priority == ctx.sampling_priority
            trace, sampled = ctx.get()
            assert sampled is True, 'priority has no impact on sampled status'

    def test_current_span(self):
        # it should return the current active span
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        assert span == ctx.get_current_span()

    def test_current_root_span_none(self):
        # it should return none when there is no root span
        ctx = Context()
        assert ctx.get_current_root_span() is None

    def test_current_root_span(self):
        # it should return the current active root span
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        assert span == ctx.get_current_root_span()

    def test_close_span(self):
        # it should keep track of closed spans, moving
        # the current active to it's parent
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        ctx.close_span(span)
        assert ctx.get_current_span() is None

    def test_get_trace(self):
        # it should return the internal trace structure
        # if the context is finished
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        span.finish()
        trace, sampled = ctx.get()
        assert [span] == trace
        assert sampled is True
        # the context should be empty
        assert 0 == len(ctx._trace)
        assert ctx._current_span is None

    def test_get_trace_empty(self):
        # it should return None if the Context is not finished
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        trace, sampled = ctx.get()
        assert trace is None
        assert sampled is None

    @mock.patch('ddtrace.internal.hostname.get_hostname')
    def test_get_report_hostname_enabled(self, get_hostname):
        get_hostname.return_value = 'test-hostname'

        with self.override_global_config(dict(report_hostname=True)):
            # Create a context and add a span and finish it
            ctx = Context()
            span = Span(tracer=None, name='fake_span')
            ctx.add_span(span)
            span.finish()

            # Assert that we have not added the tag to the span yet
            assert span.get_tag(HOSTNAME_KEY) is None

            # Assert that retrieving the trace sets the tag
            trace, _ = ctx.get()
            assert trace[0].get_tag(HOSTNAME_KEY) == 'test-hostname'
            assert span.get_tag(HOSTNAME_KEY) == 'test-hostname'

    @mock.patch('ddtrace.internal.hostname.get_hostname')
    def test_get_report_hostname_disabled(self, get_hostname):
        get_hostname.return_value = 'test-hostname'

        with self.override_global_config(dict(report_hostname=False)):
            # Create a context and add a span and finish it
            ctx = Context()
            span = Span(tracer=None, name='fake_span')
            ctx.add_span(span)
            span.finish()

            # Assert that we have not added the tag to the span yet
            assert span.get_tag(HOSTNAME_KEY) is None

            # Assert that retrieving the trace does not set the tag
            trace, _ = ctx.get()
            assert trace[0].get_tag(HOSTNAME_KEY) is None
            assert span.get_tag(HOSTNAME_KEY) is None

    @mock.patch('ddtrace.internal.hostname.get_hostname')
    def test_get_report_hostname_default(self, get_hostname):
        get_hostname.return_value = 'test-hostname'

        # Create a context and add a span and finish it
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        span.finish()

        # Assert that we have not added the tag to the span yet
        assert span.get_tag(HOSTNAME_KEY) is None

        # Assert that retrieving the trace does not set the tag
        trace, _ = ctx.get()
        assert trace[0].get_tag(HOSTNAME_KEY) is None
        assert span.get_tag(HOSTNAME_KEY) is None

    def test_partial_flush(self):
        """
        When calling `Context.get`
        When partial flushing is enabled
        When we have just enough finished spans to flush
        We return the finished spans
        """
        tracer = get_dummy_tracer()
        ctx = Context()

        # Create a root span with 5 children, all of the children are finished, the root is not
        root = Span(tracer=tracer, name='root')
        ctx.add_span(root)
        for i in range(5):
            child = Span(tracer=tracer, name='child_{}'.format(i), trace_id=root.trace_id, parent_id=root.span_id)
            child._parent = root
            child.finished = True
            ctx.add_span(child)
            ctx.close_span(child)

        with self.override_partial_flush(ctx, enabled=True, min_spans=5):
            trace, sampled = ctx.get()

        self.assertIsNotNone(trace)
        self.assertIsNotNone(sampled)

        self.assertEqual(len(trace), 5)
        self.assertEqual(
            set(['child_0', 'child_1', 'child_2', 'child_3', 'child_4']),
            set([span.name for span in trace])
        )

        # Ensure we clear/reset internal stats as expected
        self.assertEqual(ctx._trace, [root])
        with self.override_partial_flush(ctx, enabled=True, min_spans=5):
            trace, sampled = ctx.get()
            self.assertIsNone(trace)
            self.assertIsNone(sampled)

    def test_partial_flush_too_many(self):
        """
        When calling `Context.get`
        When partial flushing is enabled
        When we have more than the minimum number of spans needed to flush
        We return the finished spans
        """
        tracer = get_dummy_tracer()
        ctx = Context()

        # Create a root span with 5 children, all of the children are finished, the root is not
        root = Span(tracer=tracer, name='root')
        ctx.add_span(root)
        for i in range(5):
            child = Span(tracer=tracer, name='child_{}'.format(i), trace_id=root.trace_id, parent_id=root.span_id)
            child._parent = root
            child.finished = True
            ctx.add_span(child)
            ctx.close_span(child)

        with self.override_partial_flush(ctx, enabled=True, min_spans=1):
            trace, sampled = ctx.get()

        self.assertIsNotNone(trace)
        self.assertIsNotNone(sampled)

        self.assertEqual(len(trace), 5)
        self.assertEqual(
            set(['child_0', 'child_1', 'child_2', 'child_3', 'child_4']),
            set([span.name for span in trace])
        )

        # Ensure we clear/reset internal stats as expected
        self.assertEqual(ctx._trace, [root])
        with self.override_partial_flush(ctx, enabled=True, min_spans=5):
            trace, sampled = ctx.get()
            self.assertIsNone(trace)
            self.assertIsNone(sampled)

    def test_partial_flush_too_few(self):
        """
        When calling `Context.get`
        When partial flushing is enabled
        When we do not have enough finished spans to flush
        We return no spans
        """
        tracer = get_dummy_tracer()
        ctx = Context()

        # Create a root span with 5 children, all of the children are finished, the root is not
        root = Span(tracer=tracer, name='root')
        ctx.add_span(root)
        for i in range(5):
            child = Span(tracer=tracer, name='child_{}'.format(i), trace_id=root.trace_id, parent_id=root.span_id)
            child._parent = root
            child.finished = True
            ctx.add_span(child)
            ctx.close_span(child)

        # Test with having 1 too few spans for partial flush
        with self.override_partial_flush(ctx, enabled=True, min_spans=6):
            trace, sampled = ctx.get()

        self.assertIsNone(trace)
        self.assertIsNone(sampled)

        self.assertEqual(len(ctx._trace), 6)
        self.assertEqual(
            set(['root', 'child_0', 'child_1', 'child_2', 'child_3', 'child_4']),
            set([span.name for span in ctx._trace])
        )

    def test_partial_flush_remaining(self):
        """
        When calling `Context.get`
        When partial flushing is enabled
        When we have some unfinished spans
        We keep the unfinished spans around
        """
        tracer = get_dummy_tracer()
        ctx = Context()

        # Create a root span with 5 children, all of the children are finished, the root is not
        root = Span(tracer=tracer, name='root')
        ctx.add_span(root)
        for i in range(10):
            child = Span(tracer=tracer, name='child_{}'.format(i), trace_id=root.trace_id, parent_id=root.span_id)
            child._parent = root
            ctx.add_span(child)

            # CLose the first 5 only
            if i < 5:
                child.finished = True
                ctx.close_span(child)

        with self.override_partial_flush(ctx, enabled=True, min_spans=5):
            trace, sampled = ctx.get()

        # Assert partially flushed spans
        self.assertTrue(len(trace), 5)
        self.assertIsNotNone(sampled)
        self.assertEqual(
            set(['child_0', 'child_1', 'child_2', 'child_3', 'child_4']),
            set([span.name for span in trace])
        )

        # Assert remaining unclosed spans
        self.assertEqual(len(ctx._trace), 6)
        self.assertEqual(
            set(['root', 'child_5', 'child_6', 'child_7', 'child_8', 'child_9']),
            set([span.name for span in ctx._trace]),
        )

    def test_finished(self):
        # a Context is finished if all spans inside are finished
        ctx = Context()
        span = Span(tracer=None, name='fake_span')
        ctx.add_span(span)
        ctx.close_span(span)

    @mock.patch('logging.Logger.debug')
    def test_log_unfinished_spans_disabled(self, log):
        # the trace finished status logging is disabled
        tracer = get_dummy_tracer()
        ctx = Context()
        # manually create a root-child trace
        root = Span(tracer=tracer, name='root')
        child_1 = Span(tracer=tracer, name='child_1', trace_id=root.trace_id, parent_id=root.span_id)
        child_2 = Span(tracer=tracer, name='child_2', trace_id=root.trace_id, parent_id=root.span_id)
        child_1._parent = root
        child_2._parent = root
        ctx.add_span(root)
        ctx.add_span(child_1)
        ctx.add_span(child_2)
        # close only the parent
        root.finish()
        # the logger has never been invoked to print unfinished spans
        for call, _ in log.call_args_list:
            msg = call[0]
            assert 'the trace has %d unfinished spans' not in msg

    @mock.patch('logging.Logger.debug')
    def test_log_unfinished_spans_when_ok(self, log):
        # if the unfinished spans logging is enabled but the trace is finished, don't log anything
        tracer = get_dummy_tracer()
        ctx = Context()
        # manually create a root-child trace
        root = Span(tracer=tracer, name='root')
        child = Span(tracer=tracer, name='child_1', trace_id=root.trace_id, parent_id=root.span_id)
        child._parent = root
        ctx.add_span(root)
        ctx.add_span(child)
        # close the trace
        child.finish()
        root.finish()
        # the logger has never been invoked to print unfinished spans
        for call, _ in log.call_args_list:
            msg = call[0]
            assert 'the trace has %d unfinished spans' not in msg

    def test_thread_safe(self):
        # the Context must be thread-safe
        ctx = Context()

        def _fill_ctx():
            span = Span(tracer=None, name='fake_span')
            ctx.add_span(span)

        threads = [threading.Thread(target=_fill_ctx) for _ in range(100)]

        for t in threads:
            t.daemon = True
            t.start()

        for t in threads:
            t.join()

        assert 100 == len(ctx._trace)

    def test_clone(self):
        ctx = Context()
        ctx.sampling_priority = 2
        # manually create a root-child trace
        root = Span(tracer=None, name='root')
        child = Span(tracer=None, name='child_1', trace_id=root.trace_id, parent_id=root.span_id)
        child._parent = root
        ctx.add_span(root)
        ctx.add_span(child)
        cloned_ctx = ctx.clone()
        assert cloned_ctx._parent_trace_id == ctx._parent_trace_id
        assert cloned_ctx._parent_span_id == ctx._parent_span_id
        assert cloned_ctx._sampling_priority == ctx._sampling_priority
        assert cloned_ctx._dd_origin == ctx._dd_origin
        assert cloned_ctx._current_span == ctx._current_span
        assert cloned_ctx._trace == []



import importlib
import sys

from ddtrace.utils import deprecation


deprecation.deprecation("ddtrace.profile", "Use ddtrace.profiling instead.")

sys.modules[__name__] = importlib.import_module(__name__.replace("ddtrace.profile", "ddtrace.profiling"))



import os
import time

import tornado.web
import tornado.concurrent

from . import uimodules
from .compat import sleep, ThreadPoolExecutor


BASE_DIR = os.path.dirname(os.path.realpath(__file__))
STATIC_DIR = os.path.join(BASE_DIR, 'statics')


class SuccessHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        self.write('OK')


class NestedHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        tracer = self.settings['datadog_trace']['tracer']
        with tracer.trace('tornado.sleep'):
            yield sleep(0.05)
        self.write('OK')


class NestedWrapHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        tracer = self.settings['datadog_trace']['tracer']

        # define a wrapped coroutine: having an inner coroutine
        # is only for easy testing
        @tracer.wrap('tornado.coro')
        @tornado.gen.coroutine
        def coro():
            yield sleep(0.05)

        yield coro()
        self.write('OK')


class NestedExceptionWrapHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        tracer = self.settings['datadog_trace']['tracer']

        # define a wrapped coroutine: having an inner coroutine
        # is only for easy testing
        @tracer.wrap('tornado.coro')
        @tornado.gen.coroutine
        def coro():
            yield sleep(0.05)
            raise Exception('Ouch!')

        yield coro()
        self.write('OK')


class ExceptionHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        raise Exception('Ouch!')


class HTTPExceptionHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        raise tornado.web.HTTPError(status_code=501, log_message='unavailable', reason='Not Implemented')


class HTTPException500Handler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        raise tornado.web.HTTPError(status_code=500, log_message='server error', reason='Server Error')


class TemplateHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        self.render('templates/page.html', name='home')


class TemplatePartialHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        self.render('templates/list.html', items=['python', 'go', 'ruby'])


class TemplateExceptionHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        self.render('templates/exception.html')


class SyncSuccessHandler(tornado.web.RequestHandler):
    def get(self):
        self.write('OK')


class SyncExceptionHandler(tornado.web.RequestHandler):
    def get(self):
        raise Exception('Ouch!')


class SyncNestedWrapHandler(tornado.web.RequestHandler):
    def get(self):
        tracer = self.settings['datadog_trace']['tracer']

        # define a wrapped coroutine: having an inner coroutine
        # is only for easy testing
        @tracer.wrap('tornado.func')
        def func():
            time.sleep(0.05)

        func()
        self.write('OK')


class SyncNestedExceptionWrapHandler(tornado.web.RequestHandler):
    def get(self):
        tracer = self.settings['datadog_trace']['tracer']

        # define a wrapped coroutine: having an inner coroutine
        # is only for easy testing
        @tracer.wrap('tornado.func')
        def func():
            time.sleep(0.05)
            raise Exception('Ouch!')

        func()
        self.write('OK')


class CustomDefaultHandler(tornado.web.ErrorHandler):
    """
    Default handler that is used in case of 404 error; in our tests
    it's used only if defined in the get_app() function.
    """
    pass


class ExecutorHandler(tornado.web.RequestHandler):
    # used automatically by the @run_on_executor decorator
    executor = ThreadPoolExecutor(max_workers=3)

    @tornado.concurrent.run_on_executor
    def outer_executor(self):
        tracer = self.settings['datadog_trace']['tracer']
        with tracer.trace('tornado.executor.with'):
            time.sleep(0.05)

    @tornado.gen.coroutine
    def get(self):
        yield self.outer_executor()
        self.write('OK')


class ExecutorSubmitHandler(tornado.web.RequestHandler):
    executor = ThreadPoolExecutor(max_workers=3)

    def query(self):
        tracer = self.settings['datadog_trace']['tracer']
        with tracer.trace('tornado.executor.query'):
            time.sleep(0.05)

    @tornado.gen.coroutine
    def get(self):
        # run the query in another Executor, without using
        # Tornado decorators
        yield self.executor.submit(self.query)
        self.write('OK')


class ExecutorDelayedHandler(tornado.web.RequestHandler):
    # used automatically by the @run_on_executor decorator
    executor = ThreadPoolExecutor(max_workers=3)

    @tornado.concurrent.run_on_executor
    def outer_executor(self):
        # waiting here means expecting that the `get()` flushes
        # the request trace
        time.sleep(0.01)
        tracer = self.settings['datadog_trace']['tracer']
        with tracer.trace('tornado.executor.with'):
            time.sleep(0.05)

    @tornado.gen.coroutine
    def get(self):
        # we don't yield here but we expect that the outer_executor
        # has the right parent; tests that use this handler, must
        # yield sleep() to wait thread execution
        self.outer_executor()
        self.write('OK')


try:
    class ExecutorCustomHandler(tornado.web.RequestHandler):
        # not used automatically, a kwarg is required
        custom_thread_pool = ThreadPoolExecutor(max_workers=3)

        @tornado.concurrent.run_on_executor(executor='custom_thread_pool')
        def outer_executor(self):
            # wait before creating a trace so that we're sure
            # the `tornado.executor.with` span has the right
            # parent
            tracer = self.settings['datadog_trace']['tracer']
            with tracer.trace('tornado.executor.with'):
                time.sleep(0.05)

        @tornado.gen.coroutine
        def get(self):
            yield self.outer_executor()
            self.write('OK')
except TypeError:
    # the class definition fails because Tornado 4.0 and 4.1 don't support
    # `run_on_executor` with params. Because it's just this case, we can
    # use a try-except block, but if we have another case we may move
    # these endpoints outside the module and use a compatibility system
    class ExecutorCustomHandler(tornado.web.RequestHandler):
        pass


class ExecutorCustomArgsHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    def get(self):
        # this is not a legit use of the decorator so a failure is expected
        @tornado.concurrent.run_on_executor(object(), executor='_pool')
        def outer_executor(self):
            pass

        yield outer_executor(self)
        self.write('OK')


class ExecutorExceptionHandler(tornado.web.RequestHandler):
    # used automatically by the @run_on_executor decorator
    executor = ThreadPoolExecutor(max_workers=3)

    @tornado.concurrent.run_on_executor
    def outer_executor(self):
        # wait before creating a trace so that we're sure
        # the `tornado.executor.with` span has the right
        # parent
        time.sleep(0.05)
        tracer = self.settings['datadog_trace']['tracer']
        with tracer.trace('tornado.executor.with'):
            raise Exception('Ouch!')

    @tornado.gen.coroutine
    def get(self):
        yield self.outer_executor()
        self.write('OK')


class ExecutorWrapHandler(tornado.web.RequestHandler):
    # used automatically by the @run_on_executor decorator
    executor = ThreadPoolExecutor(max_workers=3)

    @tornado.gen.coroutine
    def get(self):
        tracer = self.settings['datadog_trace']['tracer']

        @tracer.wrap('tornado.executor.wrap')
        @tornado.concurrent.run_on_executor
        def outer_executor(self):
            time.sleep(0.05)

        yield outer_executor(self)
        self.write('OK')


class ExecutorExceptionWrapHandler(tornado.web.RequestHandler):
    # used automatically by the @run_on_executor decorator
    executor = ThreadPoolExecutor(max_workers=3)

    @tornado.gen.coroutine
    def get(self):
        tracer = self.settings['datadog_trace']['tracer']

        @tracer.wrap('tornado.executor.wrap')
        @tornado.concurrent.run_on_executor
        def outer_executor(self):
            time.sleep(0.05)
            raise Exception('Ouch!')

        yield outer_executor(self)
        self.write('OK')


def make_app(settings={}):
    """
    Create a Tornado web application, useful to test
    different behaviors.
    """
    settings['ui_modules'] = uimodules

    return tornado.web.Application([
        # custom handlers
        (r'/success/', SuccessHandler),
        (r'/nested/', NestedHandler),
        (r'/nested_wrap/', NestedWrapHandler),
        (r'/nested_exception_wrap/', NestedExceptionWrapHandler),
        (r'/exception/', ExceptionHandler),
        (r'/http_exception/', HTTPExceptionHandler),
        (r'/http_exception_500/', HTTPException500Handler),
        (r'/template/', TemplateHandler),
        (r'/template_partial/', TemplatePartialHandler),
        (r'/template_exception/', TemplateExceptionHandler),
        # handlers that spawn new threads
        (r'/executor_handler/', ExecutorHandler),
        (r'/executor_submit_handler/', ExecutorSubmitHandler),
        (r'/executor_delayed_handler/', ExecutorDelayedHandler),
        (r'/executor_custom_handler/', ExecutorCustomHandler),
        (r'/executor_custom_args_handler/', ExecutorCustomArgsHandler),
        (r'/executor_exception/', ExecutorExceptionHandler),
        (r'/executor_wrap_handler/', ExecutorWrapHandler),
        (r'/executor_wrap_exception/', ExecutorExceptionWrapHandler),
        # built-in handlers
        (r'/redirect/', tornado.web.RedirectHandler, {'url': '/success/'}),
        (r'/statics/(.*)', tornado.web.StaticFileHandler, {'path': STATIC_DIR}),
        # synchronous handlers
        (r'/sync_success/', SyncSuccessHandler),
        (r'/sync_exception/', SyncExceptionHandler),
        (r'/sync_nested_wrap/', SyncNestedWrapHandler),
        (r'/sync_nested_exception_wrap/', SyncNestedExceptionWrapHandler),
    ], **settings)



from falcon import testing
import falcon as falcon
from .app import get_app
from .test_suite import FalconTestCase
from ... import TracerTestCase


class MiddlewareTestCase(TracerTestCase, testing.TestCase, FalconTestCase):
    """Executes tests using the manual instrumentation so a middleware
    is explicitly added.
    """
    def setUp(self):
        super(MiddlewareTestCase, self).setUp()

        # build a test app with a dummy tracer
        self._service = 'falcon'
        self.api = get_app(tracer=self.tracer)
        self.version = falcon.__version__
        if(self.version[0] != '1'):
            self.client = testing.TestClient(self.api)



import asyncio

from ddtrace.provider import DefaultContextProvider
from .utils import AsyncioTestCase, mark_asyncio


class TestAsyncioSafety(AsyncioTestCase):
    """
    Ensure that if the ``AsyncioTracer`` is not properly configured,
    bad traces are produced but the ``Context`` object will not
    leak memory.
    """

    def setUp(self):
        # Asyncio TestCase with the wrong context provider
        super(TestAsyncioSafety, self).setUp()
        self.tracer.configure(context_provider=DefaultContextProvider())

    @mark_asyncio
    def test_get_call_context(self):
        # it should return a context even if not attached to the Task
        ctx = self.tracer.get_call_context()
        assert ctx is not None
        # test that it behaves the wrong way
        task = asyncio.Task.current_task()
        task_ctx = getattr(task, "__datadog_context", None)
        assert task_ctx is None

    @mark_asyncio
    def test_trace_coroutine(self):
        # it should use the task context when invoked in a coroutine
        with self.tracer.trace("coroutine") as span:
            span.resource = "base"

        traces = self.tracer.writer.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])
        assert "coroutine" == traces[0][0].name
        assert "base" == traces[0][0].resource

    @mark_asyncio
    def test_trace_multiple_calls(self):
        @asyncio.coroutine
        def coro():
            # another traced coroutine
            with self.tracer.trace("coroutine"):
                yield from asyncio.sleep(0.01)

        ctx = self.tracer.get_call_context()
        futures = [asyncio.ensure_future(coro()) for x in range(1000)]
        for future in futures:
            yield from future

        # the trace is wrong but the Context is finished
        traces = self.tracer.writer.pop_traces()
        assert 1 == len(traces)
        assert 1000 == len(traces[0])
        assert 0 == len(ctx._trace)



from .encoding import Encoder


class PayloadFull(Exception):
    """The payload is full."""
    pass


class Payload(object):
    """
    Trace agent API payload buffer class

    This class is used to encoded and store traces to build the payload we send to
    the trace agent.

    DEV: We encoded and buffer traces so that we can reliable determine the size of
         the payload easily so we can flush based on the payload size.
    """
    __slots__ = ('traces', 'size', 'encoder', 'max_payload_size')

    # Trace agent limit payload size of 10 MB
    # 5 MB should be a good average efficient size
    DEFAULT_MAX_PAYLOAD_SIZE = 5 * 1000000

    def __init__(self, encoder=None, max_payload_size=DEFAULT_MAX_PAYLOAD_SIZE):
        """
        Constructor for Payload

        :param encoder: The encoded to use, default is the default encoder
        :type encoder: ``ddtrace.encoding.Encoder``
        :param max_payload_size: The max number of bytes a payload should be before
            being considered full (default: 5mb)
        """
        self.max_payload_size = max_payload_size
        self.encoder = encoder or Encoder()
        self.traces = []
        self.size = 0

    def add_trace(self, trace):
        """
        Encode and append a trace to this payload

        :param trace: A trace to append
        :type trace: A list of :class:`ddtrace.span.Span`
        """
        # No trace or empty trace was given, ignore
        if not trace:
            return

        # Encode the trace, append, and add it's length to the size
        encoded = self.encoder.encode_trace(trace)
        if len(encoded) + self.size > self.max_payload_size:
            raise PayloadFull()
        self.traces.append(encoded)
        self.size += len(encoded)

    @property
    def length(self):
        """
        Get the number of traces in this payload

        :returns: The number of traces in the payload
        :rtype: int
        """
        return len(self.traces)

    @property
    def empty(self):
        """
        Whether this payload is empty or not

        :returns: Whether this payload is empty or not
        :rtype: bool
        """
        return self.length == 0

    def get_payload(self):
        """
        Get the fully encoded payload

        :returns: The fully encoded payload
        :rtype: str | bytes
        """
        # DEV: `self.traces` is an array of encoded traces, `join_encoded` joins them together
        return self.encoder.join_encoded(self.traces)

    def __repr__(self):
        """Get the string representation of this payload"""
        return '{0}(length={1}, size={2} B, max_payload_size={3} B)'.format(
            self.__class__.__name__, self.length, self.size, self.max_payload_size)



# 3p
import MySQLdb

from ddtrace.vendor.wrapt import wrap_function_wrapper as _w

# project
from ddtrace import config, Pin
from ddtrace.contrib.dbapi import TracedConnection

from ...ext import net, db
from ...utils.wrappers import unwrap as _u


config._add("mysqldb", dict(
    _default_service="mysql",
))

KWPOS_BY_TAG = {
    net.TARGET_HOST: ('host', 0),
    db.USER: ('user', 1),
    db.NAME: ('db', 3),
}


def patch():
    # patch only once
    if getattr(MySQLdb, '__datadog_patch', False):
        return
    setattr(MySQLdb, '__datadog_patch', True)

    # `Connection` and `connect` are aliases for
    # `Connect`; patch them too
    _w('MySQLdb', 'Connect', _connect)
    if hasattr(MySQLdb, 'Connection'):
        _w('MySQLdb', 'Connection', _connect)
    if hasattr(MySQLdb, 'connect'):
        _w('MySQLdb', 'connect', _connect)


def unpatch():
    if not getattr(MySQLdb, '__datadog_patch', False):
        return
    setattr(MySQLdb, '__datadog_patch', False)

    # unpatch MySQLdb
    _u(MySQLdb, 'Connect')
    if hasattr(MySQLdb, 'Connection'):
        _u(MySQLdb, 'Connection')
    if hasattr(MySQLdb, 'connect'):
        _u(MySQLdb, 'connect')


def _connect(func, instance, args, kwargs):
    conn = func(*args, **kwargs)
    return patch_conn(conn, *args, **kwargs)


def patch_conn(conn, *args, **kwargs):
    tags = {t: kwargs[k] if k in kwargs else args[p]
            for t, (k, p) in KWPOS_BY_TAG.items()
            if k in kwargs or len(args) > p}
    tags[net.TARGET_PORT] = conn.port
    pin = Pin(app='mysql', tags=tags)

    # grab the metadata from the conn
    wrapped = TracedConnection(conn, pin=pin, cfg=config.mysqldb)
    pin.onto(wrapped)
    return wrapped



import math

from ddtrace.encoding import Encoder, JSONEncoder
from ddtrace.payload import Payload, PayloadFull
from ddtrace.span import Span

from tests import TracerTestCase

import pytest


class PayloadTestCase(TracerTestCase):
    def test_init(self):
        """
        When calling `Payload.init`
            With an encoder
                We use that encoder
            With no encoder
                We use the default encoder
        """
        default_encoder_type = type(Encoder())

        payload = Payload()
        self.assertIsInstance(payload.encoder, default_encoder_type)

        json_encoder = JSONEncoder()
        payload = Payload(encoder=json_encoder)
        self.assertEqual(payload.encoder, json_encoder)

    def test_add_trace(self):
        """
        When calling `Payload.add_trace`
            With a falsey value
                Nothing is added to the payload
            With a trace
                We encode and add the trace to the payload
                We increment the payload size by the expected amount
        """
        payload = Payload()

        # Add falsey traces
        for val in (False, None, 0, '', [], dict()):
            payload.add_trace(val)
        self.assertEqual(payload.length, 0)
        self.assertTrue(payload.empty)

        # Add a single trace to the payload
        trace = [Span(self.tracer, name='root.span'), Span(self.tracer, name='child.span')]
        payload.add_trace(trace)

        self.assertEqual(payload.length, 1)
        self.assertFalse(payload.empty)

    def test_get_payload(self):
        """
        When calling `Payload.get_payload`
            With no traces
                We return the appropriate data
            With traces
                We return the appropriate data
        """
        payload = Payload()

        # No traces
        self.assertTrue(payload.empty)
        encoded_data = payload.get_payload()
        decoded_data = payload.encoder._decode(encoded_data)
        self.assertEqual(decoded_data, [])

        # Add traces to the payload
        for _ in range(5):
            trace = [Span(self.tracer, name='root.span'), Span(self.tracer, name='child.span')]
            payload.add_trace(trace)

        self.assertEqual(payload.length, 5)
        self.assertFalse(payload.empty)

        # Assert the payload generated from Payload
        encoded_data = payload.get_payload()
        decoded_data = payload.encoder._decode(encoded_data)
        self.assertEqual(len(decoded_data), 5)
        for trace in decoded_data:
            self.assertEqual(len(trace), 2)
            self.assertEqual(trace[0][b'name'], b'root.span')
            self.assertEqual(trace[1][b'name'], b'child.span')

    def test_full(self):
        payload = Payload()

        # Empty
        self.assertTrue(payload.empty)

        # Trace and it's size in bytes
        trace = [Span(self.tracer, 'root.span'), Span(self.tracer, 'child.span')]
        trace_size = len(payload.encoder.encode_trace(trace))

        # Number of traces before we hit the max size limit and are considered full
        num_traces = int(math.floor(payload.max_payload_size / trace_size))

        # Add the traces
        for _ in range(num_traces):
            payload.add_trace(trace)

        # Just confirm
        self.assertEqual(payload.length, num_traces)

        with pytest.raises(PayloadFull):
            payload.add_trace(trace)

        # Just confirm again
        self.assertEqual(payload.length, num_traces)



import os

import pytest

from ddtrace.vendor import six

from ddtrace.profiling import _line2def


def test_filename_and_lineno_to_def():
    filename = os.path.join(os.path.dirname(__file__), "_test_line2def_1.py")
    assert _line2def.filename_and_lineno_to_def(filename, 1).endswith("/_test_line2def_1.py:1")
    assert _line2def.filename_and_lineno_to_def(filename, 2).endswith("/_test_line2def_1.py:2")
    assert _line2def.filename_and_lineno_to_def(filename, 3).endswith("/_test_line2def_1.py:3")
    assert _line2def.filename_and_lineno_to_def(filename, 4).endswith("/_test_line2def_1.py:4")
    assert _line2def.filename_and_lineno_to_def(filename, 5) == "A"
    assert _line2def.filename_and_lineno_to_def(filename, 6) == "x"
    assert _line2def.filename_and_lineno_to_def(filename, 7) == "x"
    assert _line2def.filename_and_lineno_to_def(filename, 8).endswith("/_test_line2def_1.py:8")


def test_filename_and_lineno_to_def_ast():
    filename = os.path.join(os.path.dirname(__file__), "_ast_test_file.py")
    for i in range(29):
        assert _line2def.filename_and_lineno_to_def(filename, i).endswith("/_ast_test_file.py:%d" % i)
    for i in range(30, 36):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "parse"
    for i in range(36, 38):
        assert _line2def.filename_and_lineno_to_def(filename, i).endswith("/_ast_test_file.py:%d" % i)
    for i in range(38, 49):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "literal_eval"
    for i in range(49, 56):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "_convert_num"
    for i in range(56, 64):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "_convert_signed_num"
    for i in range(64, 91):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "_convert"
    for i in range(91, 92):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "literal_eval"
    for i in range(92, 94):
        assert _line2def.filename_and_lineno_to_def(filename, i).endswith("/_ast_test_file.py:%d" % i)
    for i in range(94, 103):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "dump"
    for i in range(103, 119):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "_format"
    for i in range(119, 122):
        assert _line2def.filename_and_lineno_to_def(filename, i) == "dump"


@pytest.mark.skipif(six.PY2, reason="Does not work on Python2")
def test_filename_and_lineno_to_def_async():
    filename = os.path.join(os.path.dirname(__file__), "_test_line2def_async.py")
    assert _line2def.filename_and_lineno_to_def(filename, 1).endswith("/_test_line2def_async.py:1")
    assert _line2def.filename_and_lineno_to_def(filename, 2).endswith("/_test_line2def_async.py:2")
    assert _line2def.filename_and_lineno_to_def(filename, 3).endswith("/_test_line2def_async.py:3")
    assert _line2def.filename_and_lineno_to_def(filename, 4).endswith("/_test_line2def_async.py:4")
    assert _line2def.filename_and_lineno_to_def(filename, 5) == "A"
    assert _line2def.filename_and_lineno_to_def(filename, 6) == "x"
    assert _line2def.filename_and_lineno_to_def(filename, 7) == "x"
    assert _line2def.filename_and_lineno_to_def(filename, 8).endswith("/_test_line2def_async.py:8")


def test_filename_and_lineno_to_def_oserror():
    assert _line2def.filename_and_lineno_to_def("/nonexistent", 8) == "/nonexistent:8"


def test_filename_and_lineno_to_def_wrong_syntax():
    filename = os.path.join(os.path.dirname(__file__), "_wrong_file")
    assert _line2def.filename_and_lineno_to_def(filename, 8) == ("%s:8" % filename)


def test_empty_filename_to_def():
    assert _line2def.filename_and_lineno_to_def("", 1) == ":1"


def test_bracket_filename_to_def():
    assert _line2def.filename_and_lineno_to_def("<input>", 2) == "<input>:2"
    assert _line2def.filename_and_lineno_to_def("<>", 2) == "<>:2"



import bottle
import webtest

import ddtrace
from ddtrace import compat
from ddtrace.contrib.bottle import TracePlugin

from ... import TracerTestCase, assert_span_http_status_code

SERVICE = 'bottle-app'


class TraceBottleDistributedTest(TracerTestCase):
    """
    Ensures that Bottle is properly traced.
    """
    def setUp(self):
        super(TraceBottleDistributedTest, self).setUp()

        # provide a dummy tracer
        self._original_tracer = ddtrace.tracer
        ddtrace.tracer = self.tracer
        # provide a Bottle app
        self.app = bottle.Bottle()

    def tearDown(self):
        # restore the tracer
        ddtrace.tracer = self._original_tracer

    def _trace_app_distributed(self, tracer=None):
        self.app.install(TracePlugin(service=SERVICE, tracer=tracer))
        self.app = webtest.TestApp(self.app)

    def _trace_app_not_distributed(self, tracer=None):
        self.app.install(TracePlugin(service=SERVICE, tracer=tracer, distributed_tracing=False))
        self.app = webtest.TestApp(self.app)

    def test_distributed(self):
        # setup our test app
        @self.app.route('/hi/<name>')
        def hi(name):
            return 'hi %s' % name
        self._trace_app_distributed(self.tracer)

        # make a request
        headers = {'x-datadog-trace-id': '123',
                   'x-datadog-parent-id': '456'}
        resp = self.app.get('/hi/dougie', headers=headers)
        assert resp.status_int == 200
        assert compat.to_unicode(resp.body) == u'hi dougie'

        # validate it's traced
        spans = self.tracer.writer.pop()
        assert len(spans) == 1
        s = spans[0]
        assert s.name == 'bottle.request'
        assert s.service == 'bottle-app'
        assert s.resource == 'GET /hi/<name>'
        assert_span_http_status_code(s, 200)
        assert s.get_tag('http.method') == 'GET'
        # check distributed headers
        assert 123 == s.trace_id
        assert 456 == s.parent_id

    def test_not_distributed(self):
        # setup our test app
        @self.app.route('/hi/<name>')
        def hi(name):
            return 'hi %s' % name
        self._trace_app_not_distributed(self.tracer)

        # make a request
        headers = {'x-datadog-trace-id': '123',
                   'x-datadog-parent-id': '456'}
        resp = self.app.get('/hi/dougie', headers=headers)
        assert resp.status_int == 200
        assert compat.to_unicode(resp.body) == u'hi dougie'

        # validate it's traced
        spans = self.tracer.writer.pop()
        assert len(spans) == 1
        s = spans[0]
        assert s.name == 'bottle.request'
        assert s.service == 'bottle-app'
        assert s.resource == 'GET /hi/<name>'
        assert_span_http_status_code(s, 200)
        assert s.get_tag('http.method') == 'GET'
        # check distributed headers
        assert 123 != s.trace_id
        assert 456 != s.parent_id



import sys

import mock

from ddtrace.profiling import recorder
from ddtrace.profiling.collector import exceptions


def test_collect():
    r = recorder.Recorder()
    c = exceptions.UncaughtExceptionCollector(r)
    c.start()
    c.stop()


def _check_event(event):
    assert event.thread_name == "MainThread"
    assert isinstance(event.thread_id, int)
    assert event.frames == []
    assert event.exc_type == ValueError


def test_no_override():
    r = recorder.Recorder()
    c = exceptions.UncaughtExceptionCollector(r)
    c.start()
    sys.excepthook(ValueError, ValueError(), None)
    c.stop()
    sys.excepthook(ValueError, ValueError(), None)
    events = r.events[exceptions.UncaughtExceptionEvent]
    assert len(events) == 1
    _check_event(events[0])


def test_with_override():
    seen = {"seen": False}

    def myhook(exctype, value, traceback):
        seen["seen"] = True

    sys.excepthook = myhook
    r = recorder.Recorder()
    c = exceptions.UncaughtExceptionCollector(r)
    c.start()
    sys.excepthook(ValueError, ValueError(), None)
    c.stop()
    assert seen["seen"]
    seen["seen"] = False
    sys.excepthook(ValueError, ValueError(), None)
    assert seen["seen"]
    events = r.events[exceptions.UncaughtExceptionEvent]
    assert len(events) == 1
    _check_event(events[0])


def test_call_original_on_failure():
    r = mock.Mock()

    def _raise(self):
        raise RuntimeError("oops")

    r.push_event.side_effect = _raise

    c = exceptions.UncaughtExceptionCollector(r)
    called = {"called": False}

    def set_called(exctype, value, tb):
        called["called"] = True

    orig_excepthook = sys.excepthook
    sys.excepthook = set_called

    try:
        c.start()
        c.except_hook(None, None, None)
        c.stop()
        assert called["called"]
    finally:
        sys.excepthook = orig_excepthook



import sys

from ddtrace.ext import SpanTypes, http as httpx
from ddtrace.http import store_request_headers, store_response_headers
from ddtrace.propagation.http import HTTPPropagator

from ...compat import iteritems
from ...constants import ANALYTICS_SAMPLE_RATE_KEY, SPAN_MEASURED_KEY
from ...settings import config


class TraceMiddleware(object):

    def __init__(self, tracer, service='falcon', distributed_tracing=True):
        # store tracing references
        self.tracer = tracer
        self.service = service
        self._distributed_tracing = distributed_tracing

    def process_request(self, req, resp):
        if self._distributed_tracing:
            # Falcon uppercases all header names.
            headers = dict((k.lower(), v) for k, v in iteritems(req.headers))
            propagator = HTTPPropagator()
            context = propagator.extract(headers)
            # Only activate the new context if there was a trace id extracted
            if context.trace_id:
                self.tracer.context_provider.activate(context)

        span = self.tracer.trace(
            'falcon.request',
            service=self.service,
            span_type=SpanTypes.WEB,
        )
        span.set_tag(SPAN_MEASURED_KEY)

        # set analytics sample rate with global config enabled
        span.set_tag(
            ANALYTICS_SAMPLE_RATE_KEY,
            config.falcon.get_analytics_sample_rate(use_global_config=True)
        )

        span.set_tag(httpx.METHOD, req.method)
        span.set_tag(httpx.URL, req.url)
        if config.falcon.trace_query_string:
            span.set_tag(httpx.QUERY_STRING, req.query_string)

        # Note: any request header set after this line will not be stored in the span
        store_request_headers(req.headers, span, config.falcon)

    def process_resource(self, req, resp, resource, params):
        span = self.tracer.current_span()
        if not span:
            return  # unexpected
        span.resource = '%s %s' % (req.method, _name(resource))

    def process_response(self, req, resp, resource, req_succeeded=None):
        # req_succeded is not a kwarg in the API, but we need that to support
        # Falcon 1.0 that doesn't provide this argument
        span = self.tracer.current_span()
        if not span:
            return  # unexpected

        status = httpx.normalize_status_code(resp.status)

        # Note: any response header set after this line will not be stored in the span
        store_response_headers(resp._headers, span, config.falcon)

        # FIXME[matt] falcon does not map errors or unmatched routes
        # to proper status codes, so we we have to try to infer them
        # here. See https://github.com/falconry/falcon/issues/606
        if resource is None:
            status = '404'
            span.resource = '%s 404' % req.method
            span.set_tag(httpx.STATUS_CODE, status)
            span.finish()
            return

        err_type = sys.exc_info()[0]
        if err_type is not None:
            if req_succeeded is None:
                # backward-compatibility with Falcon 1.0; any version
                # greater than 1.0 has req_succeded in [True, False]
                # TODO[manu]: drop the support at some point
                status = _detect_and_set_status_error(err_type, span)
            elif req_succeeded is False:
                # Falcon 1.1+ provides that argument that is set to False
                # if get an Exception (404 is still an exception)
                status = _detect_and_set_status_error(err_type, span)

        span.set_tag(httpx.STATUS_CODE, status)
        if 500 <= int(status) < 600:
            span.error = 1

        # Emit span hook for this response
        # DEV: Emit before closing so they can overwrite `span.resource` if they want
        config.falcon.hooks.emit('request', span, req, resp)

        # Close the span
        span.finish()


def _is_404(err_type):
    return 'HTTPNotFound' in err_type.__name__


def _detect_and_set_status_error(err_type, span):
    """Detect the HTTP status code from the current stacktrace and
    set the traceback to the given Span
    """
    if not _is_404(err_type):
        span.set_traceback()
        return '500'
    elif _is_404(err_type):
        return '404'


def _name(r):
    return '%s.%s' % (r.__module__, r.__class__.__name__)



from ddtrace.vendor import wrapt
import mysql.connector

from ddtrace import config, Pin
from ddtrace.contrib.dbapi import TracedConnection
from ...ext import net, db


config._add("mysql", dict(
    _default_service="mysql",
))

CONN_ATTR_BY_TAG = {
    net.TARGET_HOST: 'server_host',
    net.TARGET_PORT: 'server_port',
    db.USER: 'user',
    db.NAME: 'database',
}


def patch():
    wrapt.wrap_function_wrapper('mysql.connector', 'connect', _connect)
    # `Connect` is an alias for `connect`, patch it too
    if hasattr(mysql.connector, 'Connect'):
        mysql.connector.Connect = mysql.connector.connect


def unpatch():
    if isinstance(mysql.connector.connect, wrapt.ObjectProxy):
        mysql.connector.connect = mysql.connector.connect.__wrapped__
        if hasattr(mysql.connector, 'Connect'):
            mysql.connector.Connect = mysql.connector.connect


def _connect(func, instance, args, kwargs):
    conn = func(*args, **kwargs)
    return patch_conn(conn)


def patch_conn(conn):

    tags = {t: getattr(conn, a) for t, a in CONN_ATTR_BY_TAG.items() if getattr(conn, a, '') != ''}
    pin = Pin(app='mysql', tags=tags)

    # grab the metadata from the conn
    wrapped = TracedConnection(conn, pin=pin, cfg=config.mysql)
    pin.onto(wrapped)
    return wrapped



from ddtrace.contrib.pymysql import get_traced_pymysql_connection
from tests.tracer.test_tracer import get_dummy_tracer
from tests.contrib import config


def test_pre_v4():
    tracer = get_dummy_tracer()
    MySQL = get_traced_pymysql_connection(tracer, service='my-mysql-server')
    conn = MySQL(**config.MYSQL_CONFIG)
    cursor = conn.cursor()
    cursor.execute('SELECT 1')
    assert cursor.fetchone()[0] == 1



import os
import time

try:
    import tracemalloc
except ImportError:
    tracemalloc = None

import pytest

import ddtrace.profiling
from ddtrace.profiling import recorder
from ddtrace.profiling.collector import memory
from ddtrace.profiling.collector import stack

from . import test_collector


def _alloc():
    return [object() for _ in range(16384)]


@pytest.mark.skipif(tracemalloc is None, reason="tracemalloc is unavailable")
def test_collect():
    test_collector._test_collector_collect(
        memory.MemoryCollector, memory.MemorySampleEvent, fn=_alloc, capture_pct=100, ignore_profiler=False
    )


@pytest.mark.skipif(tracemalloc is None, reason="tracemalloc is unavailable")
def test_restart():
    test_collector._test_restart(memory.MemoryCollector)


@pytest.mark.skipif(tracemalloc is None, reason="tracemalloc is unavailable")
def test_repr():
    test_collector._test_repr(
        memory.MemoryCollector,
        "MemoryCollector(status=<ServiceStatus.STOPPED: 'stopped'>, "
        "recorder=Recorder(default_max_events=32768, max_events={}), "
        "capture_pct=2.0, nframes=64, ignore_profiler=True)",
    )


def test_max_capture():
    r = recorder.Recorder()
    with pytest.raises(ValueError):
        memory.MemoryCollector(r, capture_pct=-1)


def test_max_capture_over():
    r = recorder.Recorder()
    with pytest.raises(ValueError):
        memory.MemoryCollector(r, capture_pct=200)


def _test_memory_ignore(ignore):
    r = recorder.Recorder()
    # Start a stack collector so it allocates memory
    with stack.StackCollector(r) as sc:
        r = recorder.Recorder()
        c = memory.MemoryCollector(r, ignore_profiler=ignore, capture_pct=100)
        with c as mc:
            while not r.events[memory.MemorySampleEvent]:
                _ = object()
                # Allow gevent to switch to the memory collector thread
                time.sleep(0)
    sc.join()
    mc.join()
    events = r.events[memory.MemorySampleEvent]
    files = {frame.filename for event in events for trace in event.snapshot.traces for frame in trace.traceback}
    return files


@pytest.mark.skipif(tracemalloc is None, reason="tracemalloc is unavailable")
def test_memory_ignore_true():
    files = _test_memory_ignore(True)
    ddprofile_dir = os.path.dirname(ddtrace.profiling.__file__)
    assert all(map(lambda f: not f.startswith(ddprofile_dir), files)), files


@pytest.mark.skipif(tracemalloc is None, reason="tracemalloc is unavailable")
def test_memory_ignore_false():
    files = _test_memory_ignore(False)
    ddprofile_dir = os.path.dirname(ddtrace.profiling.__file__)
    assert any(map(lambda f: f.startswith(ddprofile_dir), files)), files


@pytest.mark.benchmark(group="memory-capture", min_time=1)
@pytest.mark.parametrize(
    "pct",
    range(10, 101, 10),
)
@pytest.mark.skipif(tracemalloc is None, reason="tracemalloc is unavailable")
def test_memory_alloc_speed_patched(benchmark, pct):
    r = recorder.Recorder()
    with memory.MemoryCollector(r, capture_pct=pct):
        benchmark(_alloc)


@pytest.mark.benchmark(
    group="memory-capture",
)
def test_memory_alloc_speed(benchmark):
    benchmark(_alloc)
